from typing import Any, Optional
from typing import List

from rich.progress import Progress, TextColumn, BarColumn, TaskID, MofNCompleteColumn, SpinnerColumn


class HandleProgressBar(object):
    """Generic listener for handling a progress bar."""
    __slots__ = ['message', 'progress_bar', 'task']

    def __init__(self, message: str) -> None:
        self.message: str = message
        self.progress_bar: Optional[Progress] = None
        self.task: Optional[TaskID] = None

    def init(self, items: List[Any]) -> None:
        """Initialize a progress bar on the items.

        Args:
            items (List[Any]): List of items on which init the progress bar.

        Returns:
            None
        """
        self.progress_bar = Progress(
            TextColumn("[progress.description]{task.description}", justify='left'),
            SpinnerColumn(),
            BarColumn(bar_width=None, complete_style="default", finished_style="green"),
            MofNCompleteColumn(),
            expand=True,
        )

        self.progress_bar.start()

        self.task = self.progress_bar.add_task(f"[bold][{self.message}]", total=len(items))

    def update(self, item: Any) -> None:
        """Update the progress bar with the item.

        Args:
            item (Any): Item that triggered the progress bar update.

        Returns:
            None
        """
        if self.progress_bar:
            self.progress_bar.update(self.task, advance=1, update=True)

    def finish(self) -> None:
        """Closes a progress bar.

        Returns:
            None
        """
        if self.progress_bar:
            self.progress_bar.stop_task(self.task)
            self.progress_bar.stop()

            self.task = None
            self.progress_bar = None

    def unregister(self) -> None:
        """Method called when the event is unregistered

        Returns:
            None
        """
        self.finish()



import json
import logging
import os.path
from typing import Callable, Any, Dict

from docker import DockerClient
from docker.errors import NotFound

from ... import utils
from ...exceptions import DockerPluginError
from ...os.Networking import Networking
from ...setting.Setting import Setting

LINUX_PLUGIN_NAME = "kathara/katharanp"
VDE_PLUGIN_NAME = "kathara/katharanp_vde"

HOSTTMP_KEY = "tmp"
XTABLES_CONFIGURATION_KEY = "xtables_lock"
XTABLES_LOCK_PATH = "/run/xtables.lock"


class DockerPlugin(object):
    """Class responsible for interacting with Docker Plugins."""
    __slots__ = ['client', 'current_name']

    PLUGIN_STATE_PATH = "/run/docker/runtime-runc/plugins.moby/{id}/state.json"

    def __init__(self, client: DockerClient):
        self.client: DockerClient = client
        self.current_name: str = f"{Setting.get_instance().network_plugin}:{utils.get_architecture()}"

    def check_and_download_plugin(self) -> None:
        """Check the presence of the Kathara Network Plugin and download it or upgrade it, if needed.

        Returns:
            None

        Raises:
            DockerPluginError: If the Kathara Network Plugin is not found on remote Docker connection.
            DockerPluginError: If the Kathara Network Plugin is not enabled on remote Docker connection.
        """
        try:
            logging.debug("Checking plugin `%s`..." % self.current_name)
            plugin = self.client.plugins.get(self.current_name)

            # Check for plugin updates.
            plugin.upgrade()
        except NotFound:
            if Setting.get_instance().remote_url is None:
                logging.info(f"Installing Kathara Network Plugin ({self.current_name})...")
                plugin = self.client.plugins.install(self.current_name)
                logging.info("Kathara Network Plugin installed successfully!")
            else:
                raise DockerPluginError("Kathara Network Plugin not found on remote Docker connection.")

        if Setting.get_instance().remote_url is None:
            if self.is_vde() and not plugin.enabled:
                logging.debug("Enabling plugin `%s`..." % self.current_name)
                plugin.enable()
            elif self.is_bridge():
                xtables_lock_mount = self._xtables_lock_mount()
                if not plugin.enabled:
                    self._configure_xtables_mount(plugin, xtables_lock_mount)
                    logging.debug("Enabling plugin `%s`..." % self.current_name)
                    plugin.enable()
                else:
                    # Get the mount of xtables.lock from the current plugin configuration
                    mount_obj = list(filter(lambda x: x["Name"] == XTABLES_CONFIGURATION_KEY,
                                            plugin.attrs["Settings"]["Mounts"])).pop()

                    # If it's not equal to the computed one, fix the mount
                    if mount_obj["Source"] != xtables_lock_mount:
                        plugin.disable()
                        self._configure_xtables_mount(plugin, xtables_lock_mount)
                        logging.debug("Enabling plugin `%s`..." % self.current_name)
                        plugin.enable()
        else:
            if not plugin.enabled:
                raise DockerPluginError("Kathara Network Plugin not enabled on remote Docker connection.")

    @staticmethod
    def is_vde() -> bool:
        """Check if the current enabled plugin is the VDE version.

        Returns:
            bool: True if the running plugin is the VDE version.
        """
        return Setting.get_instance().network_plugin == VDE_PLUGIN_NAME

    @staticmethod
    def is_bridge() -> bool:
        """Check if the current enabled plugin is the Linux bridge version.

        Returns:
            bool: True if the running plugin is the Linux bridge version.
        """
        return Setting.get_instance().network_plugin == LINUX_PLUGIN_NAME

    def exec_by_version(self, fun_vde: Callable, fun_bridge: Callable) -> Any:
        """Executes the callback depending on the enabled plugin version.

        Returns:
            Any: The result of the callback.
        """
        if self.is_vde():
            return fun_vde()
        elif self.is_bridge():
            return fun_bridge()

    def plugin_pid(self) -> int:
        """Get the plugin process PID from the plugin state file.

        Returns:
            int: The plugin process PID.
        """
        state = self._get_plugin_state()
        return state['init_process_pid']

    def plugin_store_path(self) -> str:
        """Get the plugin storage path (VDE only) from the plugin settings.

        Returns:
            str: The plugin storage path.

        Raises:
            FileNotFoundError: If the storage path mount point cannot be found.
        """
        plugin = self.client.plugins.get(self.current_name)
        settings = plugin.settings

        hosttmp_mount = None
        for mount in settings['Mounts']:
            if mount['Name'] == HOSTTMP_KEY:
                hosttmp_mount = mount['Destination']
                break

        if hosttmp_mount:
            return os.path.join(hosttmp_mount, "katharanp")

        raise FileNotFoundError(f"Unable to find `{HOSTTMP_KEY}` in plugin mounts.")

    def _get_plugin_state(self) -> Dict:
        """Get the plugin state.json file content from the Docker plugin state path.

        Returns:
            Dict: The content of the state.json file, parsed. Empty dict if the file cannot be found.
        """
        plugin = self.client.plugins.get(self.current_name)
        plugin_state_json = self.PLUGIN_STATE_PATH.format(id=plugin.id)
        if not os.path.exists(plugin_state_json):
            return {}
        with open(plugin_state_json, "r") as state_file:
            return json.loads(state_file.read())

    def _xtables_lock_mount(self) -> str:
        """Get the xtables.lock path depending on the host iptables version (Linux bridge only).

        Returns:
            str: The xtables.lock path.
        """

        def _mount_xtables_lock_linux():
            iptables_version = Networking.get_iptables_version()
            return "" if 'nf_tables' in iptables_version else XTABLES_LOCK_PATH

        def _mount_xtables_lock_windows():
            docker_info = self.client.info()
            return "" if 'microsoft' not in docker_info['KernelVersion'] else XTABLES_LOCK_PATH

        return utils.exec_by_platform(_mount_xtables_lock_linux, _mount_xtables_lock_windows, lambda: "")

    @staticmethod
    def _configure_xtables_mount(plugin, xtables_lock_mount) -> None:
        """Changes the Docker plugin configuration by settings the correct xtables.lock path (Linux bridge only).

        Returns:
            None
        """
        logging.debug("Configuring xtables.lock source to `%s`..." % xtables_lock_mount)
        plugin.configure({
            XTABLES_CONFIGURATION_KEY + '.source': xtables_lock_mount
        })



def strtobool(val: str) -> bool:
    """Convert a string representation of truth to true or false.

    Args:
        val (str): The value to convert. True values are 'y', 'yes', 't', 'true', 'on', and '1';
            false values are 'n', 'no', 'f', 'false', 'off', and '0'.

        Returns:
            bool: The value converted to bool.

        Raises:
            ValueError: If the specified value is not a valid representation of truth.
    """
    val = val.lower()
    if val in ('y', 'yes', 't', 'true', 'on', '1'):
        return True
    elif val in ('n', 'no', 'f', 'false', 'off', '0'):
        return False
    else:
        raise ValueError(f"Invalid truth value `{val}`.")



import argparse
from typing import List, Optional

from rich.live import Live

from ..ui.utils import create_lab_table
from ... import utils
from ...exceptions import PrivilegeError
from ...foundation.cli.command.Command import Command
from ...manager.Kathara import Kathara
from ...strings import strings, wiki_description


class ListCommand(Command):
    def __init__(self) -> None:
        Command.__init__(self)

        self.parser: argparse.ArgumentParser = argparse.ArgumentParser(
            prog='kathara list',
            description=strings['list'],
            epilog=wiki_description,
            add_help=False
        )

        self.parser.add_argument(
            '-h', '--help',
            action='help',
            default=argparse.SUPPRESS,
            help='Show a help message and exit.'
        )

        self.parser.add_argument(
            '-a', '--all',
            required=False,
            action='store_true',
            help='Show all running Kathara devices of all users. MUST BE ROOT FOR THIS OPTION.'
        )

        self.parser.add_argument(
            '-w', '-l', '--watch', '--live',
            required=False,
            action='store_true',
            help='Watch mode.'
        )

        self.parser.add_argument(
            '-n', '--name',
            metavar='DEVICE_NAME',
            required=False,
            help='Show only information about a specified device.'
        )

    def run(self, current_path: str, argv: List[str]) -> int:
        self.parse_args(argv)
        args = self.get_args()

        if args['all'] and not utils.is_admin():
            raise PrivilegeError("You must be root in order to show all Kathara devices of all users.")

        all_users = bool(args['all'])

        if args['watch']:
            self._get_live_info(machine_name=args['name'], all_users=all_users)
        else:
            with self.console.status(
                    f"Loading...",
                    spinner="dots"
            ) as _:
                machines_stats = Kathara.get_instance().get_machines_stats(
                    machine_name=args['name'], all_users=all_users
                )
                self.console.print(create_lab_table(machines_stats))

        return 0

    def _get_live_info(self, machine_name: Optional[str], all_users: bool) -> None:
        machines_stats = Kathara.get_instance().get_machines_stats(machine_name=machine_name, all_users=all_users)
        with Live(None, refresh_per_second=12.5, screen=True) as live:
            live.update(self.console.status(f"Loading...", spinner="dots"))
            live.refresh_per_second = 1
            while True:
                table = create_lab_table(machines_stats)
                if not table:
                    break

                live.update(table)



import os
import sys

from ..utils import open_machine_terminal
from .... import utils
from ....model import Machine as MachinePackage
from ....setting.Setting import Setting


class HandleMachineTerminal(object):
    """Listener fired when a device is deployed and started."""

    def run(self, item: 'MachinePackage.Machine') -> None:
        """If allowed, open a terminal, with the emulator specified in the settings, into the device.

        Args:
            item (Kathara.model.Machine): Device where open a terminal.

        Returns:
            None
        """
        if Setting.get_instance().open_terminals:
            for i in range(0, item.get_num_terms()):
                open_machine_terminal(item)

    def flush(self) -> None:
        """Clean the stdout buffer.

        Returns:
            None
        """

        def flush_unix():
            sys.stdout.write("\033[2J")
            sys.stdout.write("\033[0;0H")
            sys.stdout.flush()

        def flush_win():
            os.system('cls')

        utils.exec_by_platform(flush_unix, flush_win, flush_unix)

    def print_wait_msg(self) -> None:
        """Print the startup commands waiting message.

        Returns:
            None
        """
        sys.stdout.write("Waiting startup commands execution. Press [ENTER] to override...")
        sys.stdout.flush()



import argparse
import os
import platform
import sys
from typing import List

from ..ui.utils import create_panel
from ... import utils
from ... import version
from ...foundation.cli.command.Command import Command
from ...manager.Kathara import Kathara
from ...model.Lab import Lab
from ...setting.Setting import Setting
from ...strings import strings, wiki_description


class CheckCommand(Command):
    def __init__(self) -> None:
        Command.__init__(self)

        self.parser: argparse.ArgumentParser = argparse.ArgumentParser(
            prog='kathara check',
            description=strings['check'],
            epilog=wiki_description,
            add_help=False
        )

        self.parser.add_argument(
            '-h', '--help',
            action='help',
            default=argparse.SUPPRESS,
            help='Show a help message and exit.'
        )

    def run(self, current_path: str, argv: List[str]) -> int:
        self.parse_args(argv)

        self.console.print(create_panel("System Check", style="blue bold", justify="center"))
        self.console.print(
            f"Current Manager is:\t\t[bold dark_orange3]{Kathara.get_instance().get_formatted_manager_name()}"
        )
        self.console.print(
            f"Manager version is:\t\t[bold dark_orange3]{Kathara.get_instance().get_release_version()}"
        )
        self.console.print(
            "Python version is:\t\t[bold dark_orange3]%s" % sys.version.replace("\n", "- ")
        )
        self.console.print(
            f"Kathara version is:\t\t[bold dark_orange3]{version.CURRENT_VERSION}"
        )

        def linux_platform_info():
            info = os.uname()
            return "%s-%s-%s" % (info.sysname, info.release, info.machine)

        platform_info = utils.exec_by_platform(
            linux_platform_info, lambda: platform.platform(), lambda: platform.platform()
        )
        self.console.print(f"Operating System version is:\t[bold dark_orange3]{str(platform_info)}")

        with self.console.status(
                f"Trying to run container with `{Setting.get_instance().image}` image...",
                spinner="dots"
        ) as _:
            Setting.get_instance().open_terminals = False

            lab = Lab("kathara_test")
            lab.add_option('hosthome_mount', False)

            machine = lab.get_or_new_machine("hello_world")
            try:
                Kathara.get_instance().deploy_machine(machine)
                Kathara.get_instance().undeploy_machine(machine)
                self.console.print("[bold green]\u2713 Container run successfully.")
            except Exception as e:
                self.console.print(f"[bold red]\u00d7 Running container failed: {str(e)}")
                return 1

        return 0



from typing import List, BinaryIO, TextIO

from fs.base import FS

from .FilesystemMixin import FilesystemMixin
from ...model import Machine as MachinePackage


class LabFilesystemMixin(FilesystemMixin):
    """Abstraction to manage filesystems of network scenarios.

    Attributes:
        fs (FS): An object referencing a filesystem. Can be both real OS or a memory fs.
    """

    def __init__(self):
        super().__init__()

    def create_startup_file_from_string(self, machine: 'MachinePackage.Machine', commands_string: str) -> None:
        """Create the startup file for the specified device from a string.

        Args:
            machine (Kathara.model.Machine.Machine): The device to create the startup file for.
            commands_string (str): The startup commands for the device.

        Returns:
            None

        Raises:
            InvocationError: If the fs is None.
        """
        self.create_file_from_string(commands_string, f"{machine.name}.startup")

    def create_startup_file_from_list(self, machine: 'MachinePackage.Machine', commands: List[str]) -> None:
        """Create the startup file for the specified device from a list of strings.

        Args:
            machine (Kathara.model.Machine.Machine): The device to create the startup file for.
            commands (str): The startup commands for the device.

        Returns:
            None

        Raises:
            InvocationError: If the fs is None.
        """
        self.create_file_from_list(commands, f"{machine.name}.startup")

    def create_startup_file_from_path(self, machine: 'MachinePackage.Machine', src_path: str) -> None:
        """Create the startup file for the specified device from an existing file on the host filesystem.

        Args:
            machine (Kathara.model.Machine.Machine): The device to create the startup file for.
            src_path (str): The path of the file on the host filesystem to copy.

        Returns:
            None

        Raises:
            InvocationError: If the fs is None.
        """
        self.create_file_from_path(src_path, f"{machine.name}.startup")

    def create_startup_file_from_stream(self, machine: 'MachinePackage.Machine', stream: BinaryIO | TextIO) -> None:
        """Create the startup file for a device from a stream.

        Args:
            machine (Kathara.model.Machine.Machine): The device to create the startup file for.
            stream (Union[BinaryIO, TextIO]): The stream representing the content of the file to create.

        Returns:
            None

        Raises:
            InvocationError: If the fs is None.
            UnsupportedOperation: If the stream is opened without read permissions.
        """
        self.create_file_from_stream(stream, f"{machine.name}.startup")

    def update_startup_file_from_string(self, machine: 'MachinePackage.Machine', commands_string: str) -> None:
        """Append the command_string to the startup file for the specified device.

        Args:
            machine (Kathara.model.Machine.Machine): The device to update the startup file for.
            commands_string (str): The startup commands to add to the device.

        Returns:
            None

        Raises:
            InvocationError: If the fs is None.
        """
        self.update_file_from_string(commands_string, f"{machine.name}.startup")

    def update_startup_file_from_list(self, machine: 'MachinePackage.Machine', commands: List[str]) -> None:
        """Append the commands to the startup file for the specified device.

        Args:
            machine (Kathara.model.Machine.Machine): The device to update the startup file for.
            commands (str): The startup commands to append to the device.

        Returns:
            None

        Raises:
            InvocationError: If the fs is None.
        """
        self.update_file_from_list(commands, f"{machine.name}.startup")



from collections.abc import Iterator

from ....foundation.manager.exec_stream.IExecStream import IExecStream


class KubernetesExecStream(IExecStream):
    """Kubernetes-specific class for handling the commands stream exec.

    Attributes:
        _stream (Generator): The generator yielding the output of the stream exec.
        _stream_api_object (Any): The specific API object to interact with the underlying stream.
    """

    def stream_next(self) -> Iterator:
        """Return the next element from the stream.

        Returns:
            Iterator: The output iterator from the stream.
        """
        return next(self._stream)

    def exit_code(self) -> int:
        """Return the exit code of the execution.

        Returns:
            int: The exit code of the execution.
        """
        return int(self._stream_api_object.returncode)



import argparse
import sys
from typing import List

import chardet

from ... import utils
from ...foundation.cli.command.Command import Command
from ...manager.Kathara import Kathara
from ...model.Lab import Lab
from ...parser.netkit.LabParser import LabParser
from ...strings import strings, wiki_description


class ExecCommand(Command):
    def __init__(self) -> None:
        Command.__init__(self)

        self.parser: argparse.ArgumentParser = argparse.ArgumentParser(
            prog='kathara exec',
            description=strings['exec'],
            epilog=wiki_description,
            add_help=False
        )

        self.parser.add_argument(
            '-h', '--help',
            action='help',
            default=argparse.SUPPRESS,
            help='Show a help message and exit.'
        )

        group = self.parser.add_mutually_exclusive_group(required=False)

        group.add_argument(
            '-d', '--directory',
            help='Specify the folder containing the network scenario.',
        )
        group.add_argument(
            '-v', '--vmachine',
            dest="vmachine",
            action="store_true",
            help='The device has been started with vstart command.',
        )
        self.parser.add_argument(
            '--no-stdout',
            dest="no_stdout",
            action="store_true",
            help='Disable stdout of the executed command.',
        )
        self.parser.add_argument(
            '--no-stderr',
            dest="no_stderr",
            action="store_true",
            help='Disable stderr of the executed command.',
        )
        self.parser.add_argument(
            '--wait',
            dest="wait",
            action="store_true",
            default=False,
            help='Wait until startup commands execution finishes.',
        )
        self.parser.add_argument(
            'machine_name',
            metavar='DEVICE_NAME',
            help='Name of the device to execute the command into.'
        )
        self.parser.add_argument(
            'command',
            metavar='COMMAND',
            nargs='+',
            help='Shell command that will be executed inside the device.'
        )

    def run(self, current_path: str, argv: List[str]) -> int:
        self.parse_args(argv)
        args = self.get_args()

        if args['vmachine']:
            lab = Lab("kathara_vlab")
        else:
            lab_path = args['directory'].replace('"', '').replace("'", '') if args['directory'] else current_path
            lab_path = utils.get_absolute_path(lab_path)

            # Load custom 'kathara.conf' if it exists
            self._load_custom_configuration(lab_path)

            try:
                lab = LabParser.parse(lab_path)
            except (Exception, IOError):
                lab = Lab(None, path=lab_path)

        exec_output = Kathara.get_instance().exec(
            args['machine_name'],
            args['command'] if len(args['command']) > 1 else args['command'].pop(),
            lab_hash=lab.hash,
            wait=args['wait']
        )

        try:
            while True:
                (stdout, stderr) = next(exec_output)

                if not args['no_stdout']:
                    stdout_char_encoding = chardet.detect(stdout) if stdout else None
                    stdout = stdout.decode(stdout_char_encoding['encoding']) if stdout else ""
                    sys.stdout.write(stdout)
                if stderr and not args['no_stderr']:
                    stderr_char_encoding = chardet.detect(stderr) if stderr else None
                    stderr = stderr.decode(stderr_char_encoding['encoding']) if stderr else ""
                    sys.stderr.write(stderr)
        except StopIteration:
            pass

        return exec_output.exit_code()



import logging
import mmap
import os
import re

from ...model.Lab import Lab, LAB_METADATA
from ...utils import parse_cd_mac_address, RESERVED_MACHINE_NAMES


class LabParser(object):
    """Class responsible for parsing the lab.conf file."""

    @staticmethod
    def parse(path: str, conf_name: str = "lab.conf") -> Lab:
        """Parse the lab configuration identified by conf_name and return the corresponding Kathara network scenario.

        Args:
            path (str): The path to the directory containing the configuration file.
            conf_name (str): The name of the network scenario configuration file (default is 'lab.conf').

        Returns:
            Kathara.model.Lab.Lab: A Kathara network scenario.
        """
        lab_conf_path = os.path.join(path, conf_name)

        if not os.path.exists(lab_conf_path):
            raise IOError(f"No {conf_name} in given directory.")

        if os.stat(lab_conf_path).st_size == 0:
            raise IOError(f"{conf_name} file is empty.")

        # Reads lab.conf in memory, so it is faster.
        try:
            with open(lab_conf_path, 'r') as lab_file:
                lab_mem_file = mmap.mmap(lab_file.fileno(), 0, access=mmap.ACCESS_READ)
        except Exception:
            raise IOError(f"Cannot open {conf_name} file.")

        lab = Lab(None, path=path)

        line_number = 1
        line = lab_mem_file.readline().decode('utf-8')
        while line:
            matches = re.search(
                r"^(?P<key>[a-z0-9_]{1,30})\[(?P<arg>\w+)\]=([\"\']?)(?P<value>[^\"\']+)(\3)(\s+\#.*)?$",
                line.strip()
            )

            if matches:
                key = matches.group("key").strip()
                arg = matches.group("arg").strip()
                value = matches.group("value").replace('"', '').replace("'", '')

                if key in RESERVED_MACHINE_NAMES:
                    raise ValueError(f"In {conf_name} - Line {line_number}: "
                                     f"`{key}` is a reserved name, you can not use it for a device.")

                try:
                    # It's an interface, handle it.
                    interface_number = int(arg)

                    try:
                        cd_name, mac_address = parse_cd_mac_address(value)
                    except SyntaxError as e:
                        raise SyntaxError(f"In {conf_name} - Line {line_number}: {str(e)}")

                    if re.search(r"^\w+$", cd_name):
                        lab.connect_machine_to_link(key, cd_name,
                                                    machine_iface_number=interface_number, mac_address=mac_address)
                    else:
                        raise SyntaxError(f"In {conf_name} - Line {line_number}: "
                                          f"Collision domain `{value}` contains non-alphanumeric characters.")
                except ValueError:
                    # Not an interface, add it to the machine metas.
                    if lab.assign_meta_to_machine(key, arg, value) is not None:
                        logging.warning(f"In {conf_name} - Line {line_number}: "
                                        f"Device `{key}` already has a value assigned to meta `{arg}`. "
                                        f"Previous value has been overwritten with `{value}`.")
            else:
                if not line.startswith('#') and \
                        line.strip():
                    if not any([line.startswith(f"{x}=") for x in LAB_METADATA]):
                        raise SyntaxError(f"In {conf_name} - Line {line_number}: `{line}`.")
                    else:
                        (key, value) = line.split("=")
                        key = key.replace("LAB_", "").lower()
                        setattr(lab, key, value.replace('"', '').replace("'", '').strip())

            line_number += 1
            line = lab_mem_file.readline().decode('utf-8')

        lab.check_integrity()

        return lab



import collections
import logging
import os
import re
from io import BytesIO
from typing import Dict, Any, Tuple, Optional, List, OrderedDict, TextIO, Union, BinaryIO

# noinspection PyUnresolvedReferences
from fs._bulk import Copier
from fs.base import FS
from fs.copy import copy_fs, copy_file
from fs.tarfs import WriteTarFS
# noinspection PyUnresolvedReferences
from fs.tempfs import TempFS
from fs.walk import Walker

from . import Interface as InterfacePackage
from . import Lab as LabPackage
from . import Link as LinkPackage
from .. import utils
from ..exceptions import NonSequentialMachineInterfaceError, MachineOptionError, MachineCollisionDomainError, \
    MountDeniedError
from ..foundation.model.FilesystemMixin import FilesystemMixin
from ..setting.Setting import Setting
from ..trdparty.strtobool.strtobool import strtobool

MACHINE_CAPABILITIES: List[str] = ["NET_ADMIN", "NET_RAW", "NET_BROADCAST", "NET_BIND_SERVICE", "SYS_ADMIN"]
ALLOWED_VOLUME_MODES: List[str] = ["ro", "rw", "rx"]


class Machine(FilesystemMixin):
    """A Kathara device.

    Contains information about the device and the API object to interact with the Manager.

    Attributes:
        lab (Kathara.model.Lab): The Kathara network Scenario of the device.
        name (str): The name of the device.
        interfaces (collection.OrderedDict[int, Kathara.model.Link]): A list of the collision domains of the device.
        meta (Dict[str, Any]): Keys are meta properties name, values are meta properties values.
        api_object (Any): To interact with the current Kathara Manager.
        fs (fs.FS): The filesystem of the device. Contains files and configurations associated to it.
    """
    __slots__ = ['lab', 'name', 'interfaces', 'meta', 'api_object']

    def __init__(self, lab: 'LabPackage.Lab', name: str, **kwargs) -> None:
        """Create a new instance of a Kathara device.

        Args:
            lab (Kathara.model.Lab): The Kathara network scenario of the new device.
            name (str): The name of the device.
            **kwargs: Specifies the optional parameters of the device.

        Returns:
            None
        """
        super().__init__()

        name = name.strip()
        matches = re.search(r"^[a-z0-9_]{1,30}$", name)
        if not matches:
            raise SyntaxError(f"Invalid device name `{name}`.")

        self.lab: LabPackage.Lab = lab
        self.name: str = name

        self.interfaces: OrderedDict[int, 'InterfacePackage.Interface'] = collections.OrderedDict()

        self.meta: Dict[str, Any] = {
            'exec_commands': [],
            'sysctls': {},
            'envs': {},
            'ports': {},
            'ulimits': {},
            'volumes': {}
        }

        self.api_object: Any = None

        self.fs: FS = self.lab.fs.opendir(self.name) \
            if self.lab.fs.exists(self.name) and self.lab.fs.isdir(self.name) else None

        self.update_meta(kwargs)

    def add_interface(self, link: 'LinkPackage.Link', number: int = None, mac_address: str = None) \
            -> 'InterfacePackage.Interface':
        """Add an interface to the device attached to the specified collision domain.

        Args:
            link (Kathara.model.Link): The Kathara collision domain to attach.
            number (int): The number of the new interface. If it is None, the first free number is selected.
            mac_address (str): The MAC address of the interface. If None, a generated MAC address
                is associated when the Machine is started.

        Returns:
            Interface: The object associated to this interface.

        Raises:
            MachineCollisionDomainConflictError: If the interface number specified is already used on the device.
            MachineCollisionDomainConflictError: If the device is already connected to the collision domain.
        """
        if number is None:
            number = len(self.interfaces.keys())

        if number in self.interfaces:
            raise MachineCollisionDomainError(f"Interface {number} already set on device `{self.name}`.")

        if self.name in link.machines:
            raise MachineCollisionDomainError(
                f"Device `{self.name}` is already connected to collision domain `{link.name}`."
            )

        interface = InterfacePackage.Interface(self, link, number, mac_address)
        self.interfaces[number] = interface
        link.machines[self.name] = self

        return interface

    def remove_interface(self, link: 'LinkPackage.Link') -> None:
        """Disconnect the device from the specified collision domain.

        Args:
            link (Kathara.model.Link): The Kathara collision domain to disconnect.

        Returns:
            None

        Raises:
            MachineCollisionDomainConflictError: If the device is not connected to the collision domain.
        """
        if self.name not in link.machines:
            raise MachineCollisionDomainError(
                f"Device `{self.name}` is not connected to collision domain `{link.name}`."
            )

        self.interfaces = collections.OrderedDict(
            map(
                lambda x: x if x[1] is not None and x[1].link.name != link.name else (x[0], None),
                self.interfaces.items()
            )
        )
        link.machines.pop(self.name)

    def add_meta(self, name: str, value: Any) -> None:
        """Add a meta property to the device.

        Args:
            name (str): The name of the property.
            value (Any): The value of the property.

        Returns:
            Optional[Any]: Previous value if meta was already assigned, None otherwise.

        Raises:
            MachineOptionError: If the specified value is not valid for the specified property.
        """
        if name == "privileged":
            old_value = self.meta[name] if name in self.meta else None
            self.meta[name] = strtobool(str(value))
            return old_value

        if name == "exec":
            self.meta['exec_commands'].append(value)
            return None

        if name == "bridged":
            old_value = self.meta[name] if name in self.meta else None
            self.meta[name] = strtobool(str(value))
            return old_value

        if name == "sysctl":
            matches = re.search(r"^(?P<key>net\.([\w-]+\.)+[\w-]+)=(?P<value>[^=]*)$", value)

            # Check for valid kv-pair
            if matches:
                key = matches.group("key").strip()
                val = matches.group("value").strip()

                old_value = self.meta['sysctls'][key] if key in self.meta['sysctls'] else None

                # Convert to int if possible
                self.meta['sysctls'][key] = int(val) if val.lstrip('-').isnumeric() else val
            else:
                raise MachineOptionError(
                    "Invalid sysctl value (`%s`) on `%s`, missing `=` or value not in `net.` namespace."
                    % (value, self.name)
                )
            return old_value

        if name == "env":
            matches = re.search(r"^(?P<key>\w+)=(?P<value>.*)$", value)

            # Check for valid kv-pair
            if matches:
                key = matches.group("key").strip()
                val = matches.group("value").strip()

                old_value = self.meta['envs'][key] if key in self.meta['envs'] else None

                self.meta['envs'][key] = val
            else:
                raise MachineOptionError("Invalid env value (`%s`) on `%s`." % (value, self.name))
            return old_value

        if name == "ulimit":
            # Regex to match ulimits of the form key=soft:hard
            matches = re.search(r"^(?P<key>\w+)=(?P<soft>-?\d+)(:(?P<hard>-?\d+))?$", value)

            # Check for valid ulimit pair
            if matches:
                key = matches.group("key").strip()
                soft = int(matches.group("soft").strip())
                hard = matches.group("hard")

                if hard is not None:
                    hard = int(hard.strip())
                else:
                    hard = soft

                if soft < -1 or hard < -1:
                    raise MachineOptionError(f"Invalid ulimit value (`{value}`) on `{name}`. Values must be >= -1.")

                if soft == -1 and hard != -1:
                    raise MachineOptionError(
                        f"Invalid ulimit value (`{value}`) on `{name}`. "
                        f"Soft limit (-1) cannot be greater than hard limit ({hard})."
                    )

                if soft != -1 and hard == -1:  # Valid case, leave as is since hard limit is unlimited
                    pass
                elif soft > hard:
                    soft = hard

                old_value = self.meta['ulimits'].get(key, None)

                self.meta['ulimits'][key] = {'soft': soft, 'hard': hard}
            else:
                raise MachineOptionError(f"Invalid ulimit value (`{value}`) on `{name}`.")
            return old_value

        if name == "port":
            if '/' in value:
                (ports, protocol) = value.split('/')
            else:
                (ports, protocol) = value, 'tcp'

            if ':' not in ports:
                host_port, guest_port = 3000, ports
            else:
                (host_port, guest_port) = ports.split(':')

            protocol = protocol.lower()
            if protocol not in ['tcp', 'udp', 'sctp']:
                raise MachineOptionError("Port protocol value not valid on `%s`." % self.name)

            try:
                key_tuple = (int(host_port), protocol)

                old_value = self.meta['ports'][key_tuple] if key_tuple in self.meta['ports'] else None

                self.meta['ports'][key_tuple] = int(guest_port)
            except ValueError:
                raise MachineOptionError("Port value not valid on `%s`." % self.name)
            return old_value

        if name == "volume":
            values = list(filter(lambda x: x, value.split('|')))
            if len(values) == 3:
                host_path, guest_path, mode = values
            elif len(values) == 2:
                host_path, guest_path = values
                mode = 'ro'
            else:
                raise MachineOptionError(
                    f"The volume specified `{value}` is not in a valid format: <host_path>|<guest_path>|[<mode>]"
                )

            if mode not in ALLOWED_VOLUME_MODES:
                raise MachineOptionError(
                    f"Invalid volume mode `{mode}` on `{host_path}` mount. "
                    f"Allowed values are {', '.join(ALLOWED_VOLUME_MODES)}. "
                )

            key_value = os.path.abspath(host_path)
            old_value = self.meta['volumes'][key_value] if key_value in self.meta['volumes'] else None
            self.meta['volumes'][key_value] = {'guest_path': guest_path, 'mode': mode}
            return old_value

        old_value = self.meta[name] if name in self.meta else None
        self.meta[name] = value
        return old_value

    def update_meta(self, args: Dict[str, Any]) -> None:
        """Update the device metas from a dict.

        Args:
            args (Dict[str, Any]): Keys are the meta properties names, values are the updated meta properties values.

        Returns:
            None
        """
        if 'privileged' in args and args['privileged'] is not None and args['privileged']:
            self.add_meta("privileged", True)

        if 'exec_commands' in args and args['exec_commands'] is not None:
            for command in args['exec_commands']:
                self.add_meta("exec", command)

        if 'mem' in args and args['mem'] is not None:
            self.add_meta("mem", args['mem'])

        if 'cpus' in args and args['cpus'] is not None:
            self.add_meta("cpus", args['cpus'])

        if 'image' in args and args['image'] is not None:
            self.add_meta("image", args['image'])

        if 'bridged' in args and args['bridged'] is not None and args['bridged']:
            self.add_meta("bridged", True)

        if 'ports' in args and args['ports'] is not None:
            for port in args['ports']:
                self.add_meta("port", port)

        if 'num_terms' in args and args['num_terms'] is not None:
            self.add_meta('num_terms', args['num_terms'])

        if 'sysctls' in args and args['sysctls'] is not None:
            for sysctl in args['sysctls']:
                self.add_meta("sysctl", sysctl)

        if 'envs' in args and args['envs'] is not None:
            for envs in args['envs']:
                self.add_meta("env", envs)

        if 'ulimits' in args and args['ulimits'] is not None:
            for ulimit in args['ulimits']:
                self.add_meta("ulimit", ulimit)

        if 'ipv6' in args and args['ipv6'] is not None:
            self.add_meta("ipv6", args['ipv6'])

        if 'shell' in args and args['shell'] is not None:
            self.add_meta("shell", args['shell'])

        if 'entrypoint' in args and args['entrypoint'] is not None:
            self.add_meta("entrypoint", args['entrypoint'])

        if 'args' in args and args['args'] is not None:
            self.add_meta("args", args['args'])

        if 'volumes' in args and args['volumes'] is not None:
            for volume in args['volumes']:
                self.add_meta("volume", volume)

    def check(self) -> None:
        """Sort interfaces and check if there are missing interface numbers.

        Returns:
            None

        Raises:
            NonSequentialMachineInterfaceError: If there is a missing interface number.
        """
        logging.debug(f"Checking `{self.name}` integrity...")

        sorted_keys = list(self.interfaces.keys())
        if 'bridged_iface' in self.meta:
            sorted_keys = list(self.interfaces.keys())
            sorted_keys.append(self.meta['bridged_iface'])
        sorted_keys.sort()

        for i, num_iface in enumerate(sorted_keys):
            if num_iface != i:
                raise NonSequentialMachineInterfaceError(i, self.name)

        sorted_interfaces = sorted(self.interfaces.items(), key=lambda kv: kv[0])
        logging.debug("`%s` interfaces are %s." % (self.name, sorted_interfaces))
        self.interfaces = collections.OrderedDict(sorted_interfaces)

    def pack_data(self) -> Optional[bytes]:
        """Pack machine data into a .tar.gz file and returns the tar content as a byte array.

        While packing files, it also applies the win2linux patch in order to remove UTF-8 BOM.

        Returns:
            bytes: the tar content.
        """
        is_empty = True

        file = BytesIO()
        with WriteTarFS(file, compression="gz") as tar:
            hostlab_tar_dir = tar.makedir('hostlab')
            machine_tar_dir = hostlab_tar_dir.makedir(self.name)
            if self.fs and not self.fs.isempty(''):
                copy_fs(
                    self.fs,
                    machine_tar_dir,
                    on_copy=lambda src_fs, src_path, dst_fs, dst_path: utils.convert_win_2_linux(
                        dst_fs.getsyspath(dst_path), write=True
                    ),
                    walker=Walker(exclude=utils.EXCLUDED_FILES)
                )

                is_empty = False

            for name in [f"{self.name}.startup", f"{self.name}.shutdown", "shared.startup", "shared.shutdown"]:
                if self.lab.fs.exists(name):
                    copy_file(self.lab.fs, name, hostlab_tar_dir, name)
                    utils.convert_win_2_linux(hostlab_tar_dir.getsyspath(name), write=True)
                    is_empty = False

        if not is_empty:
            file.seek(0)
            return file.read()

        # If no machine files are found, return None.
        return None

    def is_privileged(self) -> bool:
        """Return True if the device is privileged, else return False.

        Returns:
            bool: True if the device is privileged, else False.
        """
        return self.lab.global_machine_metadata["privileged"] if "privileged" in self.lab.global_machine_metadata else \
            self.meta["privileged"] if "privileged" in self.meta else False

    def get_exec_commands(self) -> List[str]:
        """Get the device exec commands.

        Returns:
            List[str]: The list containing the additional commands.
        """
        return self.meta['exec_commands']

    def is_bridged(self) -> bool:
        """Return True if the device is bridged, else return False.

        Returns:
            bool: True if the device is bridged, else False.
        """
        return self.meta['bridged'] if "bridged" in self.meta else False

    def get_sysctls(self) -> Dict[str, Union[int, str]]:
        """Get the sysctls specified for the device.

        Returns:
            Dict[str, Union[int, str]]: Keys contain the sysctls to set, values are the values to apply.
        """
        return self.meta['sysctls']

    def get_envs(self) -> Dict[str, Union[int, str]]:
        """Get the environment variables specified for the device.

        Returns:
            Dict[str, Union[int, str]]: Keys are environment variables to set, values are the values to apply.
        """
        return self.meta['envs'] if self.meta['envs'] else {}

    def get_ulimits(self) -> Dict[str, Dict[str, int]]:
        """Get the ulimits of the device.

        Returns:
            Dict[str, Dict[str, int]]: A dictionary containing ulimits of the device.
        """
        return self.meta['ulimits'] if self.meta['ulimits'] else {}

    def get_ports(self) -> Dict[Tuple[int, str], int]:
        """Get the port mapping of the device.

        Returns:
            Dict[(int, str), int]: Keys are pairs (host port, protocol), values specifies the guest port.
        """
        return self.meta['ports']

    def get_image(self) -> str:
        """Get the image of the device, if defined in options or device meta. If not, use default one.

        Returns:
            str: The name of the device image.
        """
        return self.lab.global_machine_metadata["image"] if "image" in self.lab.global_machine_metadata else \
            self.meta["image"] if "image" in self.meta else Setting.get_instance().image

    def get_mem(self) -> str:
        """Get memory limit, if defined in options. If not, use the value from device meta. Otherwise, return None.

        Returns:
            str: The memory limit of the device.

        Raises:
            MachineOptionError: If the memory value specified is not valid.
        """
        memory = self.lab.global_machine_metadata["mem"] if "mem" in self.lab.global_machine_metadata else \
            self.meta["mem"] if "mem" in self.meta else None

        if memory:
            unit = memory[-1].lower()
            if unit not in ["b", "k", "m", "g"]:
                try:
                    return "%sm" % int(memory)
                except ValueError:
                    raise MachineOptionError("Memory value not valid on `%s`." % self.name)

            try:
                return "%s%s" % (int(memory[:-1]), unit)
            except ValueError:
                raise MachineOptionError("Memory value not valid on `%s`." % self.name)

        return memory

    def get_cpu(self, multiplier: int = 1) -> Optional[int]:
        """Get the CPU limit, multiplied by a specific multiplier.

        User should pass a float value ranging from 0 to max user CPUs.
        Try to take it from options, or device meta. Otherwise, return None.

        Args:
            multiplier (int): A numeric multiplier for the CPU limit value.

        Returns:
            Optional[int]: The CPU limit of the device.

        Raises:
            MachineOptionError: If the CPU value specified is not valid.
        """
        if "cpus" in self.lab.global_machine_metadata:
            try:
                return int(float(self.lab.global_machine_metadata["cpus"]) * multiplier)
            except ValueError:
                raise MachineOptionError("CPU value not valid on `%s`." % self.name)
        elif "cpus" in self.meta:
            try:
                return int(float(self.meta["cpus"]) * multiplier)
            except ValueError:
                raise MachineOptionError("CPU value not valid on `%s`." % self.name)

        return None

    def get_shell(self) -> str:
        """Get the custom shell specified for the device.

        Returns:
            str: The path of the custom shell specified for connecting to the device.
        """
        return self.meta['shell'] if 'shell' in self.meta else Setting.get_instance().device_shell

    def get_num_terms(self) -> int:
        """Get the number of terminals to be opened for the device.

        Returns:
            int: The number of terminals to be opened.

        Raises:
            MachineOptionError: If the terminals number value specified is not valid.
        """
        num_terms = 1

        if "num_terms" in self.lab.global_machine_metadata:
            num_terms = self.lab.global_machine_metadata['num_terms']
        elif 'num_terms' in self.meta:
            num_terms = self.meta['num_terms']

        try:
            num_terms = int(num_terms)

            if num_terms < 0:
                raise MachineOptionError("Terminals Number value on `%s` must be a positive value or zero." % self.name)
        except ValueError:
            raise MachineOptionError("Terminals Number value not valid on `%s`." % self.name)

        return num_terms

    def get_volumes(self) -> dict[str, dict[str, str]]:
        """Get the paths of the additional volumes mounted on the device.

        Returns:
            dict[str, dict[str, str]]: The paths of the additional volumes mounted on the device.
                Keys represent the paths on the host, while values represent the paths on the device.

        Raises:
            MountDeniedError: If the device cannot mount volumes.
        """
        if len(self.meta["volumes"]) > 0:
            if "_mount_volumes" in self.lab.general_options:
                can_mount = self.lab.general_options["_mount_volumes"]
            else:
                policy = Setting.get_instance().volume_mount_policy
                can_mount = policy in ["Prompt", "Always"]
        else:
            can_mount = True

        if can_mount:
            return self.meta['volumes']

        raise MountDeniedError(f"Device `{self.name}` cannot mount volumes.")

    def is_ipv6_enabled(self) -> bool:
        """Check if IPv6 is enabled on the device.

        Returns:
            bool: True if it is enabled, else False.

        Raises:
            MachineOptionError: If the IPv6 value specified is not valid.
        """
        is_v6_enabled = Setting.get_instance().enable_ipv6

        try:
            if "ipv6" in self.lab.global_machine_metadata:
                is_v6_enabled = self.lab.global_machine_metadata["ipv6"]
            elif "ipv6" in self.meta:
                is_v6_enabled = self.meta["ipv6"]

            return is_v6_enabled if type(is_v6_enabled) is bool else strtobool(is_v6_enabled)
        except ValueError:
            raise MachineOptionError("IPv6 value not valid on `%s`." % self.name)

    # Override FilesystemMixin methods to handle the condition if we want to add a file but self.fs is not set
    # In this case, we create the machine directory and assign it to self.fs before calling the actual method
    def create_file_from_string(self, content: str, dst_path: str) -> None:
        """Create a file from a string in the device fs. If fs is None, create it in the network scenario.

        Args:
            content (str): The string representing the content of the file to create.
            dst_path (str): The absolute path of the fs where create the file.

        Returns:
            None

        Raises:
            fs.errors.ResourceNotFound: If the path is not found.
        """
        if not self.fs:
            self.fs = self.lab.fs.makedir(self.name, recreate=True)

        super().create_file_from_string(content, dst_path)

    def update_file_from_string(self, content: str, dst_path: str) -> None:
        """Update a file in the fs object from a string.

        Args:
            content (str): The string representing the content for updating the file.
            dst_path (str): The absolute path on the fs of the file to update.

        Returns:
            None

        Raises:
            InvocationError: If the fs is None.
            fs.errors.ResourceNotFound: If the path is not found in the fs.
        """
        if not self.fs:
            self.fs = self.lab.fs.makedir(self.name, recreate=True)

        super().update_file_from_string(content, dst_path)

    def create_file_from_list(self, lines: List[str], dst_path: str) -> None:
        """Create a file from a list of strings in the device fs. If fs is None, create it in the network scenario.

        Args:
            lines (List[str]): The list of strings representing the content of the file to create.
            dst_path (str): The absolute path of the fs where create the file.

        Returns:
            None

        Raises:
            fs.errors.ResourceNotFound: If the path is not found.
        """
        if not self.fs:
            self.fs = self.lab.fs.makedir(self.name, recreate=True)

        super().create_file_from_list(lines, dst_path)

    def update_file_from_list(self, lines: List[str], dst_path: str) -> None:
        """Update a file in the fs object from a list of strings.

        Args:
            lines (List[str]): The list of strings representing the content for updating the file.
            dst_path (str): The absolute path on the fs of the file to upload.

        Returns:
            None

        Raises:
            InvocationError: If the fs is None.
            fs.errors.ResourceNotFound: If the path is not found in the fs.
        """
        if not self.fs:
            self.fs = self.lab.fs.makedir(self.name, recreate=True)

        super().update_file_from_list(lines, dst_path)

    def create_file_from_path(self, src_path: str, dst_path: str) -> None:
        """Create a file in the device fs from an existing file on the host filesystem. If the fs is None, create it.

        Args:
            src_path (str): The path of the file on the host filesystem to copy in the fs object.
            dst_path (str): The absolute path of the fs where create the file.

        Returns:
            None

        Raises:
            fs.errors.ResourceNotFound: If the path is not found.
        """
        if not self.fs:
            self.fs = self.lab.fs.makedir(self.name, recreate=True)

        super().create_file_from_path(src_path, dst_path)

    def create_file_from_stream(self, stream: Union[BinaryIO, TextIO], dst_path: str) -> None:
        """Create a file in the device fs from a stream. If fs is None, create it in the network scenario.

        Args:
            stream (Union[BinaryIO, TextIO]): The stream representing the content of the file to create.
            dst_path (str): The absolute path of the fs where create the file.

        Returns:
            None

        Raises:
            UnsupportedOperation: If the stream is opened without read permissions.
            fs.errors.ResourceNotFound: If the path is not found.
        """
        if not self.fs:
            self.fs = self.lab.fs.makedir(self.name, recreate=True)

        super().create_file_from_stream(stream, dst_path)

    def copy_directory_from_path(self, src_path: str, dst_path: str) -> None:
        """Copy a directory from a src_path in the host filesystem into a dst_path in the fs of the device.

        Args:
             src_path (str): The source path of the directory to copy.
             dst_path (str): The destination path on the device where to copy the directory.

        Returns:
            None
        """
        if not self.fs:
            self.fs = self.lab.fs.makedir(self.name, recreate=True)

        super().copy_directory_from_path(src_path, dst_path)

    def write_line_before(self, file_path: str, line_to_add: str, searched_line: str, first_occurrence: bool = False) \
            -> int:
        """Write a new line before a specific line in a file.

        Args:
            file_path (str): The path of the file to add the new line.
            line_to_add (str): The new line to add before the searched line.
            searched_line (str): A string or a regex representing the searched line.
            first_occurrence (bool): Inserts line only before the first occurrence. Default is False.

        Returns:
            int: Number of times the line has been added.

        Raises:
            fs.errors.FileExpected: If the path is not a file.
            fs.errors.ResourceNotFound: If the path does not exist.
            LineNotFoundError: If the searched line is not found in the file.
        """
        if not self.fs:
            self.fs = self.lab.fs.makedir(self.name, recreate=True)

        return super().write_line_before(file_path, line_to_add, searched_line, first_occurrence)

    def write_line_after(self, file_path: str, line_to_add: str, searched_line: str, first_occurrence: bool = False) \
            -> int:
        """Write a new line after a specific line in a file.

        Args:
            file_path (str): The path of the file to add the new line.
            line_to_add (str): The new line to add after the searched line.
            searched_line (str): A string or a regex representing the searched line.
            first_occurrence (bool): Inserts line only after the first occurrence. Default is False.

        Returns:
            int: Number of times the line has been added.

        Raises:
            fs.errors.FileExpected: If the path is not a file.
            fs.errors.ResourceNotFound: If the path does not exist.
            LineNotFoundError: If the searched line is not found in the file.
        """
        if not self.fs:
            self.fs = self.lab.fs.makedir(self.name, recreate=True)

        return super().write_line_after(file_path, line_to_add, searched_line, first_occurrence)

    def delete_line(self, file_path: str, line_to_delete: str, first_occurrence: bool = False) -> int:
        """Delete a specified line in a file.

        Args:
            file_path (str): The path of the file to delete the line.
            line_to_delete (str): A string or a regex representing the line to delete.
            first_occurrence (bool): Deletes only first occurrence. Default is False.

        Returns:
            int: Number of times the line has been deleted.

        Raises:
            fs.errors.FileExpected: If the path is not a file.
            fs.errors.ResourceNotFound: If the path does not exist.
            LineNotFoundError: If the searched line is not found in the file.
        """
        if not self.fs:
            self.fs = self.lab.fs.makedir(self.name, recreate=True)

        return super().delete_line(file_path, line_to_delete, first_occurrence)

    def __repr__(self) -> str:
        return "Machine(%s, %s, %s)" % (self.name, self.interfaces, self.meta)

    def __str__(self) -> str:
        formatted_machine = f"Name: {self.name}"
        formatted_machine += f"\nImage: {self.get_image()}"

        if self.interfaces:
            formatted_machine += "\nInterfaces: "
            for (iface_num, interface) in self.interfaces.items():
                if interface:
                    formatted_machine += f"\n\t- {iface_num}: {interface.link.name}"
                    if interface.mac_address:
                        formatted_machine += f" (MAC Address: {interface.mac_address})"

        if 'bridged' in self.meta:
            formatted_machine += f"\nBridged Connection: {self.meta['bridged']}"

        if self.meta["sysctls"]:
            formatted_machine += "\nSysctls:"
            for (key, value) in self.meta["sysctls"].items():
                formatted_machine += f"\n\t- {key} = {value}"

        if self.meta["ports"]:
            formatted_machine += "\nExposed Ports:"
            for (key, value) in self.meta["ports"].items():
                formatted_machine += f"\n\t- Host: {key} -> Guest: {value}"

        return formatted_machine



import os
from typing import Any, Optional

from docker import DockerClient

from .....foundation.manager.terminal.core.ITerminalSession import ITerminalSession


class DockerTTYTerminalSession(ITerminalSession):
    __slots__ = ['_exec_id', '_external_fd']

    def __init__(self, handler: Any, client: DockerClient, exec_id: str) -> None:
        super().__init__(handler, client)

        self._exec_id: str = exec_id
        self._external_fd: int = handler.fileno()

    def fileno(self) -> Optional[int]:
        """Return an OS-level file descriptor for the session, if available.

        Returns:
            Optional[int]: The file descriptor, or None if the session cannot be polled via fd-based readiness APIs.
        """
        return self._external_fd

    def read(self, n: int = 4096) -> bytes:
        """Read up to n bytes from the session output stream.

        Args:
            n (int): Maximum number of bytes to read.

        Returns:
            bytes: Data read from the session.
        """
        if self._closed:
            return b""

        try:
            return os.read(self._external_fd, n)
        except OSError:
            return b""

    def write(self, data: bytes) -> None:
        """Write bytes to the session input stream.

        Args:
            data (bytes): Data to send to the session.

        Returns:
            None
        """
        if self._closed:
            return

        os.write(self._external_fd, data)

    def resize(self, cols: int, rows: int) -> None:
        """Resize the session terminal dimensions.

        Args:
            cols (int): Terminal width in columns.
            rows (int): Terminal height in rows.

        Returns:
            None
        """
        if self._closed:
            return

        self._client.api.exec_resize(self._exec_id, height=rows, width=cols)

    def close(self) -> None:
        """Close the session and release resources.

        Returns:
            None
        """
        if self._closed:
            return

        self._closed = True

        try:
            os.close(self._external_fd)
        except OSError:
            pass



import argparse
from typing import List

from ..ui.utils import alphanumeric, cd_mac, create_panel
from ...foundation.cli.command.Command import Command
from ...manager.Kathara import Kathara
from ...model.Lab import Lab
from ...strings import strings, wiki_description


class VconfigCommand(Command):
    def __init__(self) -> None:
        Command.__init__(self)

        self.parser: argparse.ArgumentParser = argparse.ArgumentParser(
            prog='kathara vconfig',
            description=strings['vconfig'],
            epilog=wiki_description,
            add_help=False
        )

        self.parser.add_argument(
            '-h', '--help',
            action='help',
            default=argparse.SUPPRESS,
            help='Show a help message and exit.'
        )
        self.parser.add_argument(
            '-n', '--name',
            metavar='DEVICE_NAME',
            required=True,
            help='Name of the device to be connected on desired collision domains.'
        )

        group = self.parser.add_mutually_exclusive_group(required=True)

        group.add_argument(
            '--add',
            type=cd_mac,
            dest='to_add',
            metavar='CD/MAC',
            nargs='+',
            help='Specify the collision domain to add.'
        )
        group.add_argument(
            '--rm',
            type=alphanumeric,
            dest='to_remove',
            metavar='CD',
            nargs='+',
            help='Specify the collision domain to remove.'
        )

    def run(self, current_path: str, argv: List[str]) -> int:
        self.parse_args(argv)
        args = self.get_args()

        lab = Lab("kathara_vlab")
        Kathara.get_instance().update_lab_from_api(lab)

        machine_name = args['name']
        device = lab.get_machine(machine_name)
        device.api_object = Kathara.get_instance().get_machine_api_object(machine_name, lab_name=lab.name)

        self.console.print(
            create_panel(f"Updating Device `{machine_name}`", style="blue bold", justify="center")
        )

        if args['to_add']:
            for cd_name, mac_address in args['to_add']:
                self.console.print(
                    f"[green]+ Adding interface to device `{machine_name}` on collision domain `{cd_name}`" +
                    (f" with MAC Address {mac_address}" if mac_address else "") +
                    f"..."
                )
                link = lab.get_or_new_link(cd_name)
                Kathara.get_instance().connect_machine_to_link(device, link, mac_address=mac_address)

        if args['to_remove']:
            for cd_to_remove in args['to_remove']:
                self.console.print(
                    f"[red]- Removing interface on collision domain `{cd_to_remove}` from device `{machine_name}`..."
                )
                link = lab.get_link(cd_to_remove)
                link.api_object = Kathara.get_instance().get_link_api_object(cd_to_remove, lab_name=lab.name)

                Kathara.get_instance().disconnect_machine_from_link(device, link)

        return 0



from . import MenuItem


class ExternalItem(MenuItem):
    """
    A base class for items that need to do stuff on the console outside of the console menu.
    Sets the terminal back to standard mode until the action is done.
    Should probably be subclassed.
    """

    def __init__(self, text, menu=None, should_exit=False):
        # Here so Sphinx doesn't copy extraneous info from the superclass's docstring
        super(ExternalItem, self).__init__(text=text, menu=menu, should_exit=should_exit)

    def set_up(self):
        """
        This class overrides this method
        """
        self.menu.pause()
        self.menu.clear_screen()

    def clean_up(self):
        """
        This class overrides this method
        """
        self.menu.clear_screen()
        self.menu.resume()



import getpass

from collections import namedtuple
from .validators.base import BaseValidator, InvalidValidator


InputResult = namedtuple("InputResult", "input_string validation_result")


class PromptFormatter(object):
    """
    Class for formatting a text input prompt, to allow overriding the message as desired.

    Default answers will appear in [square brackets] and allow the user to return that answer by simply pressing
    the Enter button.

    If a 'Quit' option is desired, set `enable_quit` to True and provide a `quit_string` (default is 'q') and
    a `quit_message` (default is '(enter q to Quit)').

    """
    @staticmethod
    def format_prompt(prompt=None, default=None, enable_quit=False, quit_string='q',
                      quit_message='(enter q to Quit)'):
        """
        Format the message presented to the user during input prompting.

        Args:
            prompt (str): The message to ask the user.
            default (str, optional): The default answer if user does not provide explicit input.
            enable_quit (bool, optional): Flag to determine whether a Quit option will be presented.
            quit_string (str, optional): The string the user must input to quit (default is 'q').
            quit_message (str, optional): The message to the user explaining how to Quit.

        Returns:
            str: The formatted prompt string.
        """
        if prompt is None:
            return None
        prompt = prompt.rstrip()
        prompt = prompt.rstrip(':')
        if enable_quit:
            prompt = "{0} {1}".format(prompt, quit_message)
        if default:
            prompt = "{0} [{1}]".format(prompt, default)
        return "{0}: ".format(prompt)


class PromptUtils(object):
    """
    Utility class with various routines for prompting for user input.
    """

    def __init__(self, screen, prompt_formatter=None):
        """
        Creates a new instance of ConsoleUtils with the specified console. If no console was
        specified, creates a new default console using the ConsoleFactory.

        Args:
            screen (:obj:`consolemenu.screen.Screen`): The Screen instance.
            prompt_formatter (:obj:`PromptFormatter`, optional): The instance of PromptFormatter for displaying
             the prompt.
        """
        self.__screen = screen
        if prompt_formatter is None:
            prompt_formatter = PromptFormatter()
        self.__prompt_formatter = prompt_formatter

    @property
    def screen(self):
        """
        :obj:`consolemenu.screen.Screen`: The Screen instance.
        """
        return self.__screen

    def clear(self):
        """
        Clear the screen.
        """
        self.__screen.clear()

    def confirm_answer(self, answer, message=None):
        """
        Prompts the user to confirm a question with a yes/no prompt.
        If no message is specified, the default message is:  "You entered {}. Is this correct?"

        Args:
            answer (str): The answer to confirm.
            message (str, optional): Optional message if a different confirmation prompt is desired.

        Returns:
            bool: True if the user confirmed Yes, or False if user specified No.

        """
        if message is None:
            message = "\nYou entered {0}.  Is this correct?".format(answer)
        return self.prompt_for_yes_or_no(message)

    def enter_to_continue(self, message=None):
        """
        A console prompt to ask the user to 'Press [Enter] to continue'.

        Args:
            message (str, optional): A message to display in place of the default.
        """
        if message:
            message = message.rstrip() + ' '
        else:
            message = 'Press [Enter] to continue '
        self.__screen.input(message)

    def input(self, prompt=None, default=None, validators=None, enable_quit=False, quit_string='q',
              quit_message='(enter q to Quit)'):
        """
        Generic prompt the user for input.

        Args:
            prompt (str): The message to prompt the user.
            default (str, optional): The default value to suggest as an answer.
            validators (:obj:`BaseValidator`, optional): The list of validators to perform input validation.
            enable_quit (bool, optional): Specifies whether the user can cancel out of the input prompt.
            quit_string (str, optional): The string which the user must input in order to quit.
            quit_message (str, optional): The message to explain how to quit.

        Returns:
            InputResult: an InputResult tuple.

        """
        prompt = self.__prompt_formatter.format_prompt(prompt=prompt, default=default, enable_quit=enable_quit,
                                                       quit_string=quit_string, quit_message=quit_message)

        input_string = self.__screen.input(prompt=prompt)

        if enable_quit and quit_string == input_string:
            raise UserQuit

        if default is not None and input_string.strip() == '':
            input_string = default

        validation_result = self.validate_input(input_string, validators)

        return InputResult(input_string=input_string, validation_result=validation_result)

    def input_password(self, message=None):
        """
        Prompt the user for a password or other confidential data.

        This is equivalent to the input() method, but does not echo inputted characters to the screen.

        Args:
            message (str): The prompt message.

        Returns:
            str: The password provided by the user.
        """
        message = self.__prompt_formatter.format_prompt(message)
        try:
            if message:
                return getpass.getpass(message)
            else:
                return getpass.getpass()
        except BaseException:
            self.__screen.println('Warning: Unable to mask input; characters will be echoed to console')
            return self.input(message)

    def printf(self, *args):
        """
        Prints the specified arguments to the screen.

        Args:
            *args: Variable length argument list.
        """
        self.__screen.printf(*args)

    def println(self, *args):
        """
        Prints the specified arguments to the screen, followed by a newline character.

        Args:
            *args: Variable length argument list.
        """
        self.__screen.println(*args)

    def prompt_and_confirm_password(self, message):
        """
        Prompt for a password using the given message, then prompt a second time for a confirmation
        password, and verify both provided passwords match. If the passwords do not match, an error
        is displayed, "Passwords do not match", and the user must input both passwords again.

        Args:
            message (str): The prompt message.

        Returns:
            str: The password.
        """
        while True:
            pwd = self.input_password(message)
            cpwd = self.input_password("Confirm password")
            if pwd == cpwd:
                return pwd
            else:
                self.__screen.println("Passwords do not match.")

    def prompt_for_bilateral_choice(self, prompt, option1, option2):
        """
        Prompt the user for a response that must be one of the two supplied choices.

        NOTE: The user input verification is case-insensitive, but will return the original case provided
        by the given options.

        Args:
            prompt (str): The prompt to present the choices to the user.
            option1 (str): The first option.
            option2 (str): The second option.

        Returns:
            str: The choice selected by the user.

        """
        if prompt is None:
            prompt = ''
        prompt = prompt.rstrip() + ' (' + option1 + '/' + option2 + ') '
        while True:
            user_input = self.__screen.input(prompt)
            if str(user_input).lower() == option1.lower():
                return option1
            elif str(user_input).lower() == option2.lower():
                return option2

    def prompt_for_trilateral_choice(self, prompt, option1, option2, option3):
        """
        Prompt the user for a response that must be one of the three supplied choices.

        NOTE: The user input verification is case-insensitive, but will return the original case provided
        by the given options.

        Args:
            prompt (str): The prompt to present the choices to the user.
            option1 (str): The first option.
            option2 (str): The second option.
            option3 (str): The third option.

        Returns:
            str: The choice selected by the user.
        """
        if prompt is None:
            prompt = ''
        prompt = prompt.rstrip() + ' (' + option1 + '/' + option2 + '/' + option3 + ')'
        while True:
            user_input = self.__screen.input(prompt)
            if str(user_input).lower() == option1.lower():
                return option1
            elif str(user_input).lower() == option2.lower():
                return option2
            elif str(user_input).lower() == option3.lower():
                return option3

    def prompt_for_yes_or_no(self, prompt):
        """
        Prompts the user with the specified question, and expects a yes (y) or no (n)
        response, returning a boolean value representing the user's answer.

        Args:
            prompt (str): The prompt to display to the user.

        Returns:
            bool: True for yes, False for no.
        """
        user_input = self.prompt_for_bilateral_choice(prompt, 'y', 'n')
        return user_input == 'y'

    def prompt_for_numbered_choice(self, choices, title=None, prompt=">"):
        """
        Displays a numbered vertical list of choices from the provided list of strings.

        Args:
            choices (:obj:`list` of :obj:`str`): The list of choices to display.
            title (str, optional): Optional title to display above the numbered list.
            prompt (str): The prompt string. Default is ">".

        Returns:
            int: The index of selected choice.
        """
        if choices is None or len(choices) < 1:
            raise Exception('choices list must contain at least one element.')

        while True:
            self.clear()

            if title:
                self.screen.println(title + "\n")

            for i in range(0, len(choices)):
                print('   {:<4}{choice}'.format(str(i + 1) + ') ', choice=choices[i]))

            answer = self.screen.input('\n{} '.format(prompt))

            try:
                index = int(answer) - 1
                if 0 <= index < len(choices):
                    return index
            except Exception as e:
                continue

    def validate_input(self, input_string, validators):
        """
        Validate the given input string against the specified list of validators.

        Args:
            input_string (str): The input string to verify.
            validators (:obj:`list` of :obj:`BaseValidator`): The list of validators.

        Returns:
            bool: The validation result. True if the input is valid; False otherwise.

        Raises:
            InvalidValidator: If the list of validators contains an invalid BaseValidator class.
        """
        validation_result = True

        if isinstance(validators, BaseValidator):
            validators = [validators]
        elif validators is None:
            validators = []

        if isinstance(validators, list):
            validation_results = []
            for validator in validators:
                if isinstance(validator, BaseValidator):
                    validation_results.append(validator.validate(input_string=input_string))
                else:
                    raise InvalidValidator("Validator {} is not a valid validator".format(validator))

            validation_result = all(validation_results)
        else:
            raise InvalidValidator("Validator {} is not a valid validator".format(validators))

        return validation_result


class UserQuit(Exception):
    """
    Exception raised when a user chooses to Quit from an input prompt.
    """
    pass



class MenuMargins(object):
    """
    Class for menu margins. A margin is the area between the maximum specified dimensions (which is usually
    the width and height of the screen) and the menu border.

    Args:
        top (int): The top margin.
        left (int):  The left margin.
        bottom (int): The bottom margin.
        right (int): The right margin.
    """

    def __init__(self, top=1, left=2, bottom=0, right=2):
        self.__left = left
        self.__right = right
        self.__top = top
        self.__bottom = bottom

    @property
    def left(self):
        """
        The left margin.

        Returns:
            int: The left margin.

        """
        return self.__left

    @left.setter
    def left(self, left):
        self.__left = left

    @property
    def right(self):
        """
        The right margin.

        Returns:
            int: The right margin.
        """
        return self.__right

    @right.setter
    def right(self, right):
        self.__right = right

    @property
    def top(self):
        """
        The top margin.

        Returns:
            int: The top margin.
        """
        return self.__top

    @top.setter
    def top(self, top):
        self.__top = top

    @property
    def bottom(self):
        """
        The bottom margin.

        Returns:
            int: The bottom margin.
        """
        return self.__bottom

    @bottom.setter
    def bottom(self, bottom):
        self.__bottom = bottom



from . import MenuItem


class SelectionItem(MenuItem):
    """
    The item type used in :class:`consolemenu.SelectionMenu`
    """

    def __init__(self, text, index, menu=None):
        """
        :ivar int index: The index of this item in the list used to initialize the :class:`consolemenu.SelectionMenu`
        """
        super(SelectionItem, self).__init__(text=text, menu=menu, should_exit=True)
        self.index = index

    def get_return(self):
        """
        :return: The index of this item in the list of strings
        :rtype: int
        """
        return self.index



from __future__ import annotations

import logging
import os

from ..exceptions import InstantiationError


class PrivilegeHandler(object):
    __slots__ = ['user_uid', 'user_gid', 'effective_user_uid', 'effective_user_gid', '_ref']

    __instance: PrivilegeHandler = None

    @staticmethod
    def get_instance() -> PrivilegeHandler:
        if PrivilegeHandler.__instance is None:
            PrivilegeHandler()

        return PrivilegeHandler.__instance

    def __init__(self) -> None:
        if PrivilegeHandler.__instance is not None:
            raise InstantiationError("This class is a singleton!")
        else:
            try:
                self.user_uid: int = os.getuid()
                self.user_gid: int = os.getgid()

                self.effective_user_uid: int = os.geteuid()
                self.effective_user_gid: int = os.getegid()
            except AttributeError:
                pass

            self._ref = 0

            PrivilegeHandler.__instance = self

    def drop_effective_privileges(self) -> None:
        logging.debug("Called `drop_effective_privileges`...")
        try:
            os.seteuid(self.user_uid)
        except OSError:
            pass

        try:
            os.setegid(self.user_gid)
        except OSError:
            pass

    def drop_privileges(self) -> None:
        logging.debug(f"Calling `drop_privileges` with current values: "
                      f"UID={os.getuid()}, EUID={os.geteuid()}, GID={os.getgid()}, EGID={os.getegid()}...")

        logging.debug(f"Reference count is {self._ref}.")
        if self._ref > 1:
            self._ref -= 1
            logging.debug(f"Reference count is {self._ref}, exiting.")
            return

        if self._ref <= 0:
            self._ref = 0
            logging.debug(f"Reference count is {self._ref}, exiting.")
            return

        logging.debug(f"Dropping privileges to EUID={self.user_uid} and EGID={self.user_gid}...")

        self.drop_effective_privileges()

        self._ref = 0
        logging.debug(f"Reference count is reset to 0.")

        logging.debug(f"Privileges after `drop_privileges`: "
                      f"UID={os.getuid()}, EUID={os.geteuid()}, GID={os.getgid()}, EGID={os.getegid()}...")

    def raise_effective_privileges(self) -> None:
        logging.debug("Called `raise_effective_privileges`...")
        try:
            os.seteuid(self.effective_user_uid)
        except OSError:
            pass

        try:
            os.setegid(self.effective_user_gid)
        except OSError:
            pass

    def raise_privileges(self) -> None:
        logging.debug(f"Calling `raise_privileges` with current values: "
                      f"UID={os.getuid()}, EUID={os.geteuid()}, GID={os.getgid()}, EGID={os.getegid()}...")

        self._ref += 1
        logging.debug(f"Reference count is now {self._ref}.")
        if self._ref > 1:
            if os.getegid() == self.user_gid:
                logging.debug("Reference count > 1, but EGID not set for the thread. Raising privileges...")
                self.raise_effective_privileges()
            else:
                logging.debug("Reference count > 1 with privileges already correct, exiting.")
            return

        logging.debug(f"Raising privileges to EUID={self.effective_user_uid} and EGID={self.effective_user_gid}...")

        self.raise_effective_privileges()

        logging.debug(f"Privileges after `raise_privileges`: "
                      f"UID={os.getuid()}, EUID={os.geteuid()}, GID={os.getgid()}, EGID={os.getegid()}...")



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Callable


class IConsoleAdapter(ABC):
    """Interface for a platform console adapter. Implementations provide the glue between a manager event loop
    and the underlying platform's console/terminal IO mechanisms."""

    @abstractmethod
    def enter_raw(self) -> None:
        """Put the terminal into raw mode.

        Returns:
            None
        """
        raise NotImplementedError("You must implement `enter_raw` method.")

    @abstractmethod
    def exit_raw(self) -> None:
        """Restore the terminal to its previous mode (reverts "enter_raw").

        Returns:
            None
        """
        raise NotImplementedError("You must implement `exit_raw` method.")

    @abstractmethod
    def install_input_reader(self, loop: Any, on_bytes: Callable[[bytes], None], on_close: Callable[[], None]) -> None:
        """Start reading from stdin and route incoming bytes to callbacks (connected to the manager handler).

        Args:
            loop (Any): The asyncio event loop instance used to register readers/tasks.
            on_bytes (Callable[[bytes], None]): Callback to handle received input bytes.
            on_close (Callable[[], None]): Callback to handle stdin closing or errors.

        Returns:
            None
        """
        raise NotImplementedError("You must implement `install_input_reader` method.")

    @abstractmethod
    def remove_input_reader(self, loop: Any) -> None:
        """Stop reading stdin previously configured by "install_input_reader".

        Args:
            loop (Any): The asyncio event loop instance used to deregister.

        Returns:
            None
        """
        raise NotImplementedError("You must implement `remove_input_reader` method.")

    @abstractmethod
    def write_stdout(self, data: bytes) -> None:
        """Write bytes to stdout.

        Args:
            data (bytes): Bytes to write to stdout.

        Returns:
            None
        """
        raise NotImplementedError("You must implement `write_stdout` method.")

    @abstractmethod
    def watch_resize(self, loop: Any, cb: Callable[[int, int], None]) -> None:
        """Start monitoring terminal resize events.

        Args:
            loop (Any): The asyncio event loop instance used to register signal handlers or callbacks.
            cb (Callable[[int, int], None]): Callback to handle the resize.

        Returns:
            None
        """
        raise NotImplementedError("You must implement `watch_resize` method.")

    @abstractmethod
    def unwatch_resize(self, loop: Any) -> None:
        """Stop monitoring terminal resize events.

        Args:
            loop (Any): The asyncio event loop instance used to deregister signal handlers or callbacks.

        Returns:
            None
        """
        raise NotImplementedError("You must implement `unwatch_resize` method.")



from . import utils as setting_utils
from ....foundation.cli.ui.setting.OptionsHandler import OptionsHandler
from ....setting.addon.DockerSettingsAddon import DEFAULTS
from ....trdparty.consolemenu import *
from ....trdparty.consolemenu.items import *
from ....trdparty.consolemenu.validators.regex import RegexValidator
from ....types import SharedCollisionDomainsOption
from ....utils import exec_by_platform


class DockerOptionsHandler(OptionsHandler):
    def add_items(self, current_menu: ConsoleMenu, menu_formatter: MenuFormatBuilder) -> None:
        # Network Plugin Option
        network_plugin_string = "Choose Docker Network Plugin version"
        network_plugin_menu = SelectionMenu(
            strings=[],
            title=network_plugin_string,
            subtitle=setting_utils.current_string("network_plugin"),
            prologue_text="""Choose Docker Network Plugin version for collision domains.
                          
                          `kathara/katharanp` plugin is based on Linux bridges.
                          
                          `kathara/katharanp_vde` plugin is based on VDE switches.

                          Default is `%s`.""" %
                          DEFAULTS['network_plugin'],
            formatter=menu_formatter
        )

        network_plugin_menu.append_item(
            FunctionItem(
                text="kathara/katharanp",
                function=setting_utils.update_setting_value,
                args=["network_plugin", "kathara/katharanp"],
                should_exit=True
            )
        )
        network_plugin_menu.append_item(
            FunctionItem(
                text="kathara/katharanp_vde",
                function=setting_utils.update_setting_value,
                args=["network_plugin", "kathara/katharanp_vde"],
                should_exit=True
            )
        )

        network_plugin_item = SubmenuItem(network_plugin_string, network_plugin_menu, current_menu)

        # Hosthome Mount Option
        hosthome_string = "Automatically mount /hosthome on startup"
        hosthome_menu = SelectionMenu(
            strings=[],
            title=hosthome_string,
            subtitle=setting_utils.current_bool("hosthome_mount"),
            prologue_text="""The home directory of the current user is made available for """
                          """reading/writing inside the device under the special directory `/hosthome`.
                          
                          Default is %s.""" %
                          setting_utils.format_bool(DEFAULTS['hosthome_mount']),
            formatter=menu_formatter
        )

        hosthome_menu.append_item(
            FunctionItem(
                text="Yes",
                function=setting_utils.update_setting_value,
                args=["hosthome_mount", True],
                should_exit=True
            )
        )
        hosthome_menu.append_item(
            FunctionItem(
                text="No",
                function=setting_utils.update_setting_value,
                args=["hosthome_mount", False],
                should_exit=True
            )
        )

        hosthome_item = SubmenuItem(hosthome_string, hosthome_menu, current_menu)

        # Shared Mount Option
        shared_string = "Automatically mount /shared on startup"
        shared_menu = SelectionMenu(
            strings=[],
            title=shared_string,
            subtitle=setting_utils.current_bool("shared_mount"),
            prologue_text="""The shared directory inside the network scenario folder is """
                          """made available for reading/writing inside the device under the special directory `/shared`.
                          
                          Default is %s.""" %
                          setting_utils.format_bool(DEFAULTS['shared_mount']),
            formatter=menu_formatter
        )

        shared_menu.append_item(
            FunctionItem(
                text="Yes",
                function=setting_utils.update_setting_value,
                args=["shared_mount", True],
                should_exit=True
            )
        )
        shared_menu.append_item(
            FunctionItem(
                text="No",
                function=setting_utils.update_setting_value,
                args=["shared_mount", False],
                should_exit=True
            )
        )

        shared_item = SubmenuItem(shared_string, shared_menu, current_menu)

        # Image Update Policy Option
        image_update_policy_string = "Docker Image Update Policy"
        image_update_policy_menu = SelectionMenu(
            strings=[],
            title=image_update_policy_string,
            subtitle=setting_utils.current_string("image_update_policy"),
            prologue_text="""Choose the policy when a Docker image update is available for a running device.             
  
                          \tDefault is %s.""" % DEFAULTS['image_update_policy'],
            formatter=menu_formatter
        )

        image_update_policy_menu.append_item(
            FunctionItem(
                text="Prompt",
                function=setting_utils.update_setting_value,
                args=["image_update_policy", "Prompt"],
                should_exit=True
            )
        )
        image_update_policy_menu.append_item(
            FunctionItem(
                text="Always",
                function=setting_utils.update_setting_value,
                args=["image_update_policy", "Always"],
                should_exit=True
            )
        )
        image_update_policy_menu.append_item(
            FunctionItem(
                text="Never",
                function=setting_utils.update_setting_value,
                args=["image_update_policy", "Never"],
                should_exit=True
            )
        )

        image_update_policy_item = SubmenuItem(image_update_policy_string, image_update_policy_menu, current_menu)

        # Shared Collision Domains Option
        shared_cds_string = "Enable Shared Collision Domains"
        shared_cds_menu = SelectionMenu(
            strings=[],
            title=shared_cds_string,
            subtitle=setting_utils.current_enum("shared_cds", SharedCollisionDomainsOption.to_string),
            prologue_text="""This option allows sharing collision domains between network scenarios and users.
                          
                          Default is: %s.""" % SharedCollisionDomainsOption.to_string(DEFAULTS['shared_cds']),
            formatter=menu_formatter
        )

        shared_cds_menu.append_item(
            FunctionItem(
                text="Share collision domains between network scenarios",
                function=setting_utils.update_setting_value,
                args=["shared_cds", SharedCollisionDomainsOption.LABS],
                should_exit=True
            )
        )
        shared_cds_menu.append_item(
            FunctionItem(
                text="Share collision domains between users",
                function=setting_utils.update_setting_value,
                args=["shared_cds", SharedCollisionDomainsOption.USERS],
                should_exit=True
            )
        )

        shared_cds_menu.append_item(
            FunctionItem(
                text="Do not share collision domains",
                function=setting_utils.update_setting_value,
                args=["shared_cds", SharedCollisionDomainsOption.NOT_SHARED],
                should_exit=True
            )
        )

        shared_cds_item = SubmenuItem(shared_cds_string, shared_cds_menu, current_menu)

        # Remote Docker Daemon Option
        def remote_url_unix():
            remote_url_string = "Configure a remote Docker connection"
            remote_url_menu = SelectionMenu(
                strings=[],
                title=remote_url_string,
                subtitle=setting_utils.current_string("remote_url"),
                prologue_text="""You can specify a remote Docker Daemon URL.
                                            
                              Default is %s.""" % DEFAULTS['remote_url'],
                formatter=menu_formatter
            )

            remote_url_menu.append_item(
                FunctionItem(
                    text="Insert a remote Docker Daemon URL",
                    function=setting_utils.update_value,
                    args=['remote_url',
                          RegexValidator(setting_utils.URL_REGEX),
                          'Write a Docker Daemon URL '
                          '(format http[s]://<remote-url>:<remote-port>):',
                          'Docker Daemon URL is not a valid URL (remove '
                          'the trailing slash, if present)'
                          ],
                    should_exit=False
                )
            )

            remote_url_item = SubmenuItem(remote_url_string, remote_url_menu, current_menu)

            # Docker Daemon TLS Path Option
            cert_path_string = "Configure a Docker Daemon TLS Cert Path"
            cert_path_menu = SelectionMenu(
                strings=[],
                title=cert_path_string,
                subtitle=setting_utils.current_string("cert_path"),
                prologue_text="""When using a remote Docker Daemon, a TLS Cert could be required.
                              
                              Default is %s.""" % DEFAULTS['cert_path'],
                formatter=menu_formatter
            )

            cert_path_menu.append_item(
                FunctionItem(
                    text="Insert a Docker Daemon TLS Cert Path",
                    function=setting_utils.update_value,
                    args=['cert_path',
                          RegexValidator(r'^.+$'),
                          'Write a TSL Cert Path:',
                          'TLS Cert Path not valid!'
                          ],
                    should_exit=False
                )
            )
            cert_path_menu.append_item(
                FunctionItem(
                    text="Reset value to Empty String",
                    function=setting_utils.update_setting_value,
                    args=["cert_path", None],
                    should_exit=False
                )
            )

            cert_path_item = SubmenuItem(cert_path_string, cert_path_menu, remote_url_menu)
            remote_url_menu.append_item(cert_path_item)

            remote_url_menu.append_item(
                FunctionItem(
                    text="Reset remote Docker connection to default",
                    function=setting_utils.update_setting_values,
                    args=[[("remote_url", None), ("cert_path", None)]],
                    should_exit=False
                )
            )

            return remote_url_item

        platform_remote_url_item = exec_by_platform(remote_url_unix, lambda: None, remote_url_unix)

        current_menu.append_item(network_plugin_item)
        current_menu.append_item(hosthome_item)
        current_menu.append_item(shared_item)
        current_menu.append_item(image_update_policy_item)
        current_menu.append_item(shared_cds_item)
        if platform_remote_url_item:
            current_menu.append_item(platform_remote_url_item)



import argparse
from typing import List

from ..ui.utils import create_panel
from ... import utils
from ...foundation.cli.command.Command import Command
from ...manager.Kathara import Kathara
from ...model.Lab import Lab
from ...parser.netkit.LabParser import LabParser
from ...strings import strings, wiki_description


class LcleanCommand(Command):
    def __init__(self) -> None:
        Command.__init__(self)

        self.parser: argparse.ArgumentParser = argparse.ArgumentParser(
            prog='kathara lclean',
            description=strings['lclean'],
            epilog=wiki_description,
            add_help=False
        )

        self.parser.add_argument(
            '-h', '--help',
            action='help',
            default=argparse.SUPPRESS,
            help='Show a help message and exit.'
        )

        self.parser.add_argument(
            '-d', '--directory',
            required=False,
            help='Specify the folder containing the network scenario.'
        )
        self.parser.add_argument(
            '--exclude',
            dest='excluded_machines',
            metavar='DEVICE_NAME',
            nargs='+',
            default=[],
            help='Exclude specified devices from clean.'
        )
        self.parser.add_argument(
            'machine_names',
            metavar='DEVICE_NAME',
            nargs='*',
            help='Clean only specified devices.'
        )

    def run(self, current_path: str, argv: List[str]) -> int:
        self.parse_args(argv)
        args = self.get_args()

        lab_path = args['directory'].replace('"', '').replace("'", '') if args['directory'] else current_path
        lab_path = utils.get_absolute_path(lab_path)

        # Load custom 'kathara.conf' if it exists
        self._load_custom_configuration(lab_path)

        try:
            lab = LabParser.parse(lab_path)
        except (Exception, IOError):
            lab = Lab(None, path=lab_path)

        self.console.print(create_panel("Stopping Network Scenario", style="blue bold", justify="center"))

        Kathara.get_instance().undeploy_lab(
            lab_hash=lab.hash,
            selected_machines=set(args['machine_names']) if args['machine_names'] else None,
            excluded_machines=set(args['excluded_machines']) if args['excluded_machines'] else None,
        )

        return 0



import hashlib
import json
import logging
import re
from functools import partial
from multiprocessing import Manager
from multiprocessing.dummy import Pool
from typing import Dict, Optional, Set, Any, List, Generator

from kubernetes import client
from kubernetes.client.api import custom_objects_api
from kubernetes.client.rest import ApiException

from .KubernetesConfig import KubernetesConfig
from .KubernetesNamespace import KubernetesNamespace
from .stats.KubernetesLinkStats import KubernetesLinkStats
from ... import utils
from ...event.EventDispatcher import EventDispatcher
from ...exceptions import InvocationError
from ...model.Lab import Lab
from ...model.Link import Link
from ...setting.Setting import Setting

MAX_K8S_LINK_NUMBER = (1 << 24) - 10

K8S_NET_GROUP = "k8s.cni.cncf.io"
K8S_NET_VERSION = "v1"
K8S_NET_PLURAL = "network-attachment-definitions"


class KubernetesLink(object):
    """The class responsible for deploying Kathara collision domains as Kubernetes networks and interact with them."""
    __slots__ = ['client', 'kubernetes_namespace', 'seed']

    def __init__(self, kubernetes_namespace: KubernetesNamespace) -> None:
        self.client: custom_objects_api.CustomObjectsApi = custom_objects_api.CustomObjectsApi()

        self.kubernetes_namespace: KubernetesNamespace = kubernetes_namespace

        self.seed: str = KubernetesConfig.get_cluster_user()

    def deploy_links(self, lab: Lab, selected_links: Set[str] = None, excluded_links: Set[str] = None) -> None:
        """Deploy all the links contained in lab.links.

        Args:
            lab (Kathara.model.Lab.Lab): A Kathara network scenario.
            selected_links (Set[str]): A set containing the name of the collision domains to deploy.
            excluded_links (Set[str]): A set containing the name of the collision domains to exclude.

        Returns:
            None

        Raises:
            InvocationError: If both `selected_links` and `excluded_links` are specified.
        """
        if selected_links and excluded_links:
            raise InvocationError(f"You can either specify `selected_links` or `excluded_links`.")

        links = lab.links.items()
        if selected_links:
            links = {
                k: v for k, v in links if k in selected_links
            }.items()
        elif excluded_links:
            links = {
                k: v for k, v in links if k not in excluded_links
            }.items()

        if len(links) > 0:
            pool_size = utils.get_pool_size()
            items = utils.chunk_list(links, pool_size)

            EventDispatcher.get_instance().dispatch("links_deploy_started", items=links)

            with Manager() as manager:
                # Read already existing VNIs before creating new networks. This will avoid VNI collisions.
                network_ids = manager.dict({
                    network_id: 1 for network_id in self._get_existing_network_ids()
                })

                with Pool(pool_size) as links_pool:
                    for chunk in items:
                        links_pool.map(func=partial(self._deploy_link, network_ids), iterable=chunk)

            EventDispatcher.get_instance().dispatch("links_deploy_ended")

    def _deploy_link(self, network_ids: Dict, link_item: (str, Link)) -> None:
        """Deploy the Link contained in the link_item.

        Args:
            network_ids (Dict): A Dict for reserving network IDs. The Key is the network_id, value is 1 if it
                is currently used.
            link_item (Tuple[str, Link]): A tuple composed by the name of the collision domain and a Link object

        Returns:
            None
        """
        (_, link) = link_item

        network_id = self._get_unique_network_id(link.name, network_ids)
        self.create(link, network_id)

        EventDispatcher.get_instance().dispatch("link_deployed", item=link)

    def create(self, link: Link, network_id: int) -> None:
        """Create a Kubernetes Network representing the collision domain object and assign it to link.api_object.

        Args:
            link (Kathara.model.Link.Link): A Kathara collision domain.
            network_id (int): The Network ID.

        Returns:
            None
        """
        # If a network with the same name exists, return it instead of creating a new one.
        networks = self.get_links_api_objects_by_filters(lab_hash=link.lab.hash, link_name=link.name)
        if networks:
            link.api_object = networks.pop()
            return

        link.api_object = self.client.create_namespaced_custom_object(group=K8S_NET_GROUP,
                                                                      version=K8S_NET_VERSION,
                                                                      namespace=link.lab.hash,
                                                                      plural=K8S_NET_PLURAL,
                                                                      body=self._build_definition(link, network_id)
                                                                      )

        # If external is defined for a link, throw a warning.
        if link.external:
            logging.warning('External is not supported on Megalos. It will be ignored.')

    def undeploy(self, lab_hash: str, selected_links: Optional[Set[str]] = None) -> None:
        """Undeploy all the links of the scenario specified by lab_hash.

        Args:
            lab_hash (str): The hash of the network scenario to undeploy.
            selected_links (Set[str]): If specified, delete only the collision domains contained in the set.

        Returns:
            None
        """
        networks = self.get_links_api_objects_by_filters(lab_hash=lab_hash)
        if selected_links:
            networks = [item for item in networks if item["metadata"]["name"] in selected_links]

        if len(networks) > 0:
            pool_size = utils.get_pool_size()
            items = utils.chunk_list(networks, pool_size)

            EventDispatcher.get_instance().dispatch("links_undeploy_started", items=networks)

            with Pool(pool_size) as links_pool:
                for chunk in items:
                    links_pool.map(func=self._undeploy_link, iterable=chunk)

            EventDispatcher.get_instance().dispatch("links_undeploy_ended")

    def wipe(self) -> None:
        """Undeploy all the running networks of the specified user.

        Returns:
            None
        """
        networks = self.get_links_api_objects_by_filters()

        pool_size = utils.get_pool_size()
        items = utils.chunk_list(networks, pool_size)

        with Pool(pool_size) as links_pool:
            for chunk in items:
                links_pool.map(func=self._undeploy_link, iterable=chunk)

    def _undeploy_link(self, link_item: Any) -> None:
        """Undeploy a Kubernetes network.

        Args:
            link_item (Any): The Kubernetes network to undeploy.

        Returns:
            None
        """
        namespace = link_item["metadata"]["namespace"]

        try:
            self.client.delete_namespaced_custom_object(group=K8S_NET_GROUP,
                                                        version=K8S_NET_VERSION,
                                                        namespace=namespace,
                                                        plural=K8S_NET_PLURAL,
                                                        name=link_item["metadata"]["name"],
                                                        body=client.V1DeleteOptions(grace_period_seconds=0),
                                                        grace_period_seconds=0
                                                        )
        except ApiException:
            pass

        EventDispatcher.get_instance().dispatch("link_undeployed", item=link_item)

    def get_links_api_objects_by_filters(self, lab_hash: str = None, link_name: str = None) -> List[Any]:
        """Return the List of Kubernetes networks.

        Args:
            lab_hash (str): The hash of a network scenario. If specified, return only the networks in the scenario, else
                return the networks in the 'default' namespace.
            link_name (str): The name of a network. If specified, return the specified network of the scenario.

        Returns:
            List[Any]: A list of Kubernetes networks.
        """
        filters = ["app=kathara"]
        if link_name:
            filters.append(f"name={link_name}")

        # Get all Kathara namespaces if lab_hash is None
        namespaces = list(map(lambda x: x.metadata.name, self.kubernetes_namespace.get_all())) \
            if not lab_hash else [lab_hash]

        networks = []
        for namespace in namespaces:
            networks.extend(
                self.client.list_namespaced_custom_object(group=K8S_NET_GROUP,
                                                          version=K8S_NET_VERSION,
                                                          namespace=namespace,
                                                          plural=K8S_NET_PLURAL,
                                                          label_selector=",".join(filters),
                                                          timeout_seconds=9999
                                                          )["items"]
            )

        return networks

    def get_links_stats(self, lab_hash: str = None, link_name: str = None) -> \
            Generator[Dict[str, KubernetesLinkStats], None, None]:
        """Return a generator containing the Kubernetes networks' stats.

        Args:
           lab_hash (str): The hash of a network scenario. If specified, return all the stats of the networks in the
                scenario.
           link_name (str): The name of a device. If specified, return the specified network stats.

        Returns:
           Generator[Dict[str, KubernetesLinkStats], None, None]: A generator containing network names as keys and
                KubernetesLinkStats as values.
        """
        networks_stats = {}

        def load_link_stats(network):
            if network['metadata']['name'] not in networks_stats:
                networks_stats[network['metadata']['name']] = KubernetesLinkStats(network)

        while True:
            networks = self.get_links_api_objects_by_filters(lab_hash=lab_hash, link_name=link_name)
            if not networks:
                yield dict()

            pool_size = utils.get_pool_size()
            items = utils.chunk_list(networks, pool_size)
            with Pool(pool_size) as links_pool:
                for chunk in items:
                    links_pool.map(func=load_link_stats, iterable=chunk)

            networks_to_remove = []
            for network_id, network_stats in networks_stats.items():
                try:
                    network_stats.update()
                except StopIteration:
                    networks_to_remove.append(network_id)
                    continue

            for k in networks_to_remove:
                networks_stats.pop(k, None)

            yield networks_stats

    def _build_definition(self, link: Link, network_id: int) -> Dict[str, str]:
        """Return a Dict containing the network definition for Kubernetes API corresponding to link.

        Args:
            link (Kathara.model.Link.Link): A Kathara collision domain.
            network_id (int): The network id.

        Returns:
            Dict[str, str]: a Dict containing the network definition for Kubernetes API
        """
        return {
            "apiVersion": "k8s.cni.cncf.io/v1",
            "kind": "NetworkAttachmentDefinition",
            "metadata": {
                "name": self.get_network_name(link.name),
                "labels": {
                    "name": link.name,
                    "app": "kathara"
                }
            },
            "spec": {
                "config": """{
                            "cniVersion": "0.3.0",
                            "name": "%s",
                            "type": "megalos",
                            "suffix": "%s",
                            "vxlanId": %d
                        }""" % (link.name.lower(), link.lab.hash[0:6], network_id)
            }
        }

    def _get_unique_network_id(self, name: str, network_ids: Dict[int, int]) -> int:
        """Return a unique network id.

        Args:
            name (str): The name of the network.
            network_ids (Dict[int, int]): A Dict for reserving network IDs. The Key is the network_id, value is 1 if it
                is currently used.

        Returns:
            int: A network ID.

        """
        network_id = self._get_network_id(name)
        offset = 1
        while network_id in network_ids:
            network_id = self._get_network_id(name, offset)
            offset += 1

        network_ids[network_id] = 1

        return network_id

    def _get_network_id(self, name: str, offset: int = 0) -> int:
        """Return the Kubernetes network ID from a Kathara collision domain name.

        Args:
            name (str): The name of a Kathara collision domain.
            offset (int): Offset to avoid collision on the generated Kubernetes network ID.

        Returns:
            int: The Kubernetes network ID.
        """
        network_name = self.seed + name
        return (offset + int(hashlib.sha256(network_name.encode('utf-8')).hexdigest(), 16)) % MAX_K8S_LINK_NUMBER

    def _get_existing_network_ids(self) -> List[int]:
        """Retrieve already deployed Kubernetes network IDs (reading the network definition).

        Returns:
            List[int]: List with all deployed Kubernetes network IDs.
        """
        network_ids = []

        for network in self.get_links_api_objects_by_filters():
            network_config = json.loads(network['spec']['config'])
            network_ids.append(int(network_config['vxlanId']))

        return network_ids

    @staticmethod
    def get_network_name(name: str) -> str:
        """Return the name of a Kubernetes Network.

        Args:
            name (str): The name of a Kathara collision domain.

        Returns:
            str: The name of the Kubernetes Network in the format "<net_prefix>-<name><suffix>".
        """
        suffix = ''
        # Underscore is replaced with -, but to keep name uniqueness append 8 chars of hash from the original name
        if '_' in name:
            suffix = '-%s' % hashlib.md5(name.encode('utf-8', errors='ignore')).hexdigest()[:8]
            name = name.replace('_', '-')

        link_name = "%s-%s%s" % (Setting.get_instance().net_prefix, name, suffix)
        return re.sub(r'[^0-9a-z\-.]+', '', link_name.lower())



from typing import Dict, Any, Generator, Optional

from docker.errors import NotFound
from docker.models.containers import Container

from ....decorators import privileged
from ....foundation.manager.stats.IMachineStats import IMachineStats
from ....utils import human_readable_bytes


class DockerMachineStats(IMachineStats):
    """The class responsible to handle Docker Machine statistics.

    Attributes:
        machine_api_object (Container): The Docker Container associated with this statistics.
        stats (Generator[Dict[str, Any], None, None]): A generator containing dicts with the Docker statistics
        lab_hash (str): The hash identifier of the network scenario of the Docker Container.
        name (str): The name of the device.
        container_name (str): The Docker Container Name.
        user (str): The user that deployed the associated Docker Network.
        image (str): The Docker Image used for deploying the Docker Container.
        status (Optional[str]): The status of the Docker Container.
        pids (Optional[int]): The number of PIDs associated with the Docker Container.
        interfaces (str): The interfaces connected to this Docker Container.
        cpu_usage (str): The cpu usage of the Docker Container.
        mem_usage (str): The memory usage of the Docker Container.
        mem_percent (str): The memory usage of the Docker Container as a percentage.
        net_usage (str): The network usage of the Docker Container.
    """
    __slots__ = ['machine_api_object', 'stats', 'lab_hash', 'name', 'container_name', 'user', 'status', 'image',
                 'pids', 'cpu_usage', 'mem_usage', 'mem_percent', 'net_usage', '_prev_stats']

    def __init__(self, machine_api_object: Container):
        self.machine_api_object: Container = machine_api_object
        self.stats: Generator[Dict[str, Any], None, None] = machine_api_object.stats(stream=True, decode=True)
        self._prev_stats: Optional[Dict[str, Any]] = None
        # Static Information
        self.lab_hash: str = machine_api_object.labels['lab_hash']
        self.name: str = machine_api_object.labels['name']
        self.container_name: str = machine_api_object.name
        self.user: Optional[str] = machine_api_object.labels['user']
        self.image: str = machine_api_object.image.tags[0]
        # Dynamic Information
        self.status: Optional[str] = None
        self.pids: Optional[int] = None
        self.interfaces: str = "-"
        self.cpu_usage: str = "-"
        self.mem_usage: str = "- / -"
        self.mem_percent: str = "-"
        self.net_usage: str = "-"

        self.update()

    @privileged
    def update(self) -> None:
        """Update dynamic statistics with the current ones.

        Returns:
            None
        """
        updated_stats = next(self.stats)
        try:
            self.machine_api_object.reload()
        except NotFound:
            # Happens while deleting
            pass

        self.status = self.machine_api_object.status
        self.pids = updated_stats['pids_stats']['current'] if 'current' in updated_stats['pids_stats'] else 0

        networks = self.machine_api_object.attrs['NetworkSettings']['Networks']
        if 'none' in networks:
            networks.pop('none')
        if 'bridge' in networks:
            networks.pop('bridge')

        # If the device is bridged, creates a dummy entry for the bridged interface for building the interfaces string
        if 'bridged_iface' in self.machine_api_object.labels:
            networks[self.machine_api_object.labels['bridged_iface']] = {
                'DriverOpts': {
                    'kathara.iface': self.machine_api_object.labels['bridged_iface'],
                    'kathara.link': "Bridged",
                }
            }

        if networks:
            self.interfaces = ", ".join(sorted(
                [f"{v['DriverOpts']['kathara.iface']}:{v['DriverOpts']['kathara.link']}" for n, v in networks.items()]))
        else:
            self.interfaces = "-"

        if self._prev_stats:
            if "system_cpu_usage" in updated_stats["cpu_stats"] and "system_cpu_usage" in self._prev_stats["cpu_stats"]:
                cpu_delta = updated_stats["cpu_stats"]["cpu_usage"]["total_usage"] - \
                            self._prev_stats["cpu_stats"]["cpu_usage"]["total_usage"]
                system_delta = updated_stats["cpu_stats"]["system_cpu_usage"] - \
                               self._prev_stats["cpu_stats"]["system_cpu_usage"]
                cpu_usage = (cpu_delta / system_delta) * updated_stats["cpu_stats"]["online_cpus"] * 100
                self.cpu_usage = f"{cpu_usage:.2f}%"

        if "usage" in updated_stats["memory_stats"]:
            usage = updated_stats["memory_stats"]["usage"]
            limit = updated_stats["memory_stats"]["limit"]
            self.mem_usage = human_readable_bytes(usage) + " / " + human_readable_bytes(limit)
            self.mem_percent = f"{((usage / limit) * 100):.2f} %"

        if "networks" in updated_stats:
            network_stats = updated_stats["networks"] if "networks" in updated_stats else {}
            rx_bytes = sum([net_stats["rx_bytes"] for (_, net_stats) in network_stats.items()])
            tx_bytes = sum([net_stats["tx_bytes"] for (_, net_stats) in network_stats.items()])
            self.net_usage = human_readable_bytes(rx_bytes) + " / " + human_readable_bytes(tx_bytes)

        self._prev_stats = updated_stats

    def to_dict(self) -> Dict[str, Any]:
        """Transform statistics into a dict representation.

        Returns:
            Dict[str, Any]: Dict containing statistics.
        """
        return {
            "network_scenario_id": self.lab_hash,
            "name": self.name,
            "container_name": self.container_name,
            "user": self.user,
            "status": self.status,
            "image": self.image,
            "pids": self.pids,
            "cpu_usage": self.cpu_usage,
            "mem_usage": self.mem_usage,
            "mem_percent": self.mem_percent,
            "net_usage": self.net_usage,
            'interfaces': self.interfaces,
        }

    def __repr__(self) -> str:
        return str(self.to_dict())

    def __str__(self) -> str:
        """Return a formatted string with the device statistics.

        Returns:
           str: A formatted string with the device statistics
        """
        formatted_stats = f"Network Scenario ID: {self.lab_hash}\n"
        formatted_stats += f"Device Name: {self.name}\n"
        formatted_stats += f"Container Name: {self.container_name}\n"
        formatted_stats += f"Status: {self.status}\n"
        formatted_stats += f"Image: {self.image}\n"
        formatted_stats += f"PIDs: {self.pids}\n"
        formatted_stats += f"CPU Usage: {self.cpu_usage}\n"
        formatted_stats += f"Memory Usage: {self.mem_usage}\n"
        formatted_stats += f"Network Usage (DL/UL): {self.net_usage}"
        formatted_stats += f"Interfaces: {self.interfaces}\n"

        return formatted_stats



import io
import logging
from typing import Set, Dict, Generator, Tuple, List, Optional, Union

import docker
import docker.models.containers
import docker.models.networks
from docker.errors import DockerException
from requests.exceptions import ConnectionError as RequestsConnectionError

from .DockerImage import DockerImage
from .DockerLink import DockerLink
from .DockerMachine import DockerMachine
from .DockerPlugin import DockerPlugin
from .exec_stream.DockerExecStream import DockerExecStream
from .stats.DockerLinkStats import DockerLinkStats
from .stats.DockerMachineStats import DockerMachineStats
from ... import utils
from ...decorators import privileged
from ...exceptions import DockerDaemonConnectionError, LinkNotFoundError, MachineCollisionDomainError, \
    InvocationError, LabNotFoundError, MachineNotRunningError
from ...exceptions import MachineNotFoundError
from ...foundation.manager.IManager import IManager
from ...model.Lab import Lab
from ...model.Link import Link
from ...model.Machine import Machine
from ...setting.Setting import Setting
from ...types import SharedCollisionDomainsOption
from ...utils import pack_files_for_tar, import_pywintypes, \
    check_required_single_not_none_var, check_single_not_none_var

pywintypes = import_pywintypes()


def check_docker_status(method):
    """Decorator function to check if Docker daemon is running properly."""

    @privileged
    def check_docker(*args, **kw):
        # Call the constructor first
        method(*args, **kw)

        # Client is initialized after constructor call
        client = args[0].client

        # Try to ping Docker, to see if it's running and raise an exception on failure
        try:
            client.ping()
        except RequestsConnectionError as e:
            raise DockerDaemonConnectionError(str(e))
        except pywintypes.error as e:
            raise DockerDaemonConnectionError(str(e))

    return check_docker


class DockerManager(IManager):
    """The class responsible to interact between Kathara and the Docker APIs."""
    __slots__ = ['client', 'docker_image', 'docker_machine', 'docker_link']

    @check_docker_status
    def __init__(self) -> None:
        remote_url = Setting.get_instance().remote_url
        try:
            if remote_url is None:
                self.client: docker.DockerClient = docker.from_env(timeout=None, max_pool_size=utils.get_pool_size())
            else:
                tls_config = docker.tls.TLSConfig(ca_cert=Setting.get_instance().cert_path)
                self.client: docker.DockerClient = docker.DockerClient(
                    base_url=remote_url, timeout=None, max_pool_size=utils.get_pool_size(), tls=tls_config
                )
        except DockerException as e:
            raise DockerDaemonConnectionError(str(e))

        docker_plugin = DockerPlugin(self.client)
        docker_plugin.check_and_download_plugin()

        self.docker_image: DockerImage = DockerImage(self.client)

        self.docker_machine: DockerMachine = DockerMachine(self.client, self.docker_image)
        self.docker_link: DockerLink = DockerLink(self.client, docker_plugin)

    @privileged
    def deploy_machine(self, machine: Machine) -> None:
        """Deploy a Kathara device.

        Args:
            machine (Kathara.model.Machine): A Kathara machine object.

        Returns:
            None

        Raises:
            LabNotFoundError: If the specified device is not associated to any network scenario.
            PrivilegeError: If the user start the device in privileged mode without having root privileges.
            NonSequentialMachineInterfaceError: If there is a missing interface number in any device of the lab.
        """
        if not machine.lab:
            raise LabNotFoundError("Device `%s` is not associated to a network scenario." % machine.name)

        machine.check()

        self.docker_link.deploy_links(machine.lab, selected_links={x.link.name for x in machine.interfaces.values()})
        self.docker_machine.deploy_machines(machine.lab, selected_machines={machine.name})

    @privileged
    def deploy_link(self, link: Link) -> None:
        """Deploy a Kathara collision domain.

        Args:
            link (Kathara.model.Link): A Kathara collision domain object.

        Returns:
            None

        Raises:
            LabNotFoundError: If the collision domain is not associated to any network scenario.
        """
        if not link.lab:
            raise LabNotFoundError("Collision domain `%s` is not associated to a network scenario." % link.name)

        self.docker_link.deploy_links(link.lab, selected_links={link.name})

    @privileged
    def deploy_lab(self, lab: Lab, selected_machines: Optional[Set[str]] = None,
                   excluded_machines: Optional[Set[str]] = None) -> None:
        """Deploy a Kathara network scenario.

        Args:
            lab (Kathara.model.Lab): A Kathara network scenario.
            selected_machines (Optional[Set[str]]): If not None, deploy only the specified devices.
            excluded_machines (Optional[Set[str]]): If not None, exclude devices from being deployed.

        Returns:
            None

        Raises:
            NonSequentialMachineInterfaceError: If there is a missing interface number in any device of the lab.
            OSError: If any link in the network scenario is attached to external interfaces and the host OS is not LINUX.
            PrivilegeError: If the user start the network scenario in privileged mode without having root privileges.
            MachineNotFoundError: If the specified devices are not in the network scenario.
            InvocationError: If both `selected_machines` and `excluded_machines` are specified.
        """
        lab.check_integrity()

        if selected_machines and excluded_machines:
            raise InvocationError(f"You can either select or exclude devices.")

        if selected_machines and not lab.has_machines(selected_machines):
            machines_not_in_lab = selected_machines - set(lab.machines.keys())
            raise MachineNotFoundError(f"The following devices are not in the network scenario: {machines_not_in_lab}.")

        if excluded_machines and not lab.has_machines(excluded_machines):
            machines_not_in_lab = excluded_machines - set(lab.machines.keys())
            raise MachineNotFoundError(f"The following devices are not in the network scenario: {machines_not_in_lab}.")

        selected_links = None
        if selected_machines:
            selected_links = lab.get_links_from_machines(selected_machines)

        excluded_links = None
        if excluded_machines:
            # Get the links of remaining machines
            running_links = lab.get_links_from_machines(set(lab.machines.keys()) - excluded_machines)
            # Get the links of the excluded machines and get the diff with the running ones
            # The remaining are the ones to delete
            excluded_links = lab.get_links_from_machines(excluded_machines) - running_links

        # Deploy all lab links.
        self.docker_link.deploy_links(lab, selected_links=selected_links, excluded_links=excluded_links)

        # Deploy all lab machines.
        self.docker_machine.deploy_machines(
            lab, selected_machines=selected_machines, excluded_machines=excluded_machines
        )

    @privileged
    def connect_machine_to_link(self, machine: Machine, link: Link, mac_address: Optional[str] = None) -> None:
        """Create a new interface on a running Kathara device and connect it to a collision domain.

        Args:
            machine (Kathara.model.Machine): A Kathara machine object.
            link (Kathara.model.Link): A Kathara collision domain object.
            mac_address (Optional[str]): The MAC address to assign to the interface.

        Returns:
            None

        Raises:
            LabNotFoundError: If the device specified is not associated to any network scenario.
            MachineNotRunningError: If the specified device is not running.
            LabNotFoundError: If the collision domain is not associated to any network scenario.
            MachineCollisionDomainConflictError: If the device is already connected to the collision domain.
        """
        if not machine.lab:
            raise LabNotFoundError("Device `%s` is not associated to a network scenario." % machine.name)

        if not machine.api_object:
            raise MachineNotRunningError(machine.name)

        machine.api_object.reload()
        if machine.api_object.status != "running":
            raise MachineNotRunningError(machine.name)

        if not link.lab:
            raise LabNotFoundError(f"Collision domain `{link.name}` is not associated to a network scenario.")

        if machine.name in link.machines:
            raise MachineCollisionDomainError(
                f"Device `{machine.name}` is already connected to collision domain `{link.name}`."
            )

        iface_number = None
        if machine.is_bridged():
            if 'bridged_iface' not in machine.meta:
                machine.add_meta('bridged_iface', int(machine.api_object.labels['bridged_iface']))
            if not machine.interfaces or machine.meta['bridged_iface'] > max(machine.interfaces.keys()):
                iface_number = machine.meta['bridged_iface'] + 1
            else:
                iface_number = max(machine.interfaces.keys()) + 1

        interface = machine.add_interface(link, mac_address=mac_address, number=iface_number)

        self.deploy_link(link)
        self.docker_machine.connect_interface(machine, interface)

    @privileged
    def disconnect_machine_from_link(self, machine: Machine, link: Link) -> None:
        """Disconnect a running Kathara device from a collision domain.

        Args:
            machine (Kathara.model.Machine): A Kathara machine object.
            link (Kathara.model.Link): The Kathara collision domain from which disconnect the running device.

        Returns:
            None

        Raises:
            LabNotFoundError: If the device specified is not associated to any network scenario.
            MachineNotRunningError: If the specified device is not running.
            LabNotFoundError: If the collision domain is not associated to any network scenario.
            MachineCollisionDomainConflictError: If the device is not connected to the collision domain.
        """
        if not machine.lab:
            raise LabNotFoundError(f"Device `{machine.name}` is not associated to a network scenario.")

        if not machine.api_object:
            raise MachineNotRunningError(machine.name)

        machine.api_object.reload()
        if machine.api_object.status != "running":
            raise MachineNotRunningError(machine.name)

        if not link.lab:
            raise LabNotFoundError(f"Collision domain `{link.name}` is not associated to a network scenario.")

        if machine.name not in link.machines:
            raise MachineCollisionDomainError(
                f"Device `{machine.name}` is not connected to collision domain `{link.name}`."
            )

        machine.remove_interface(link)

        self.docker_machine.disconnect_from_link(machine, link)
        self.undeploy_link(link)

    @privileged
    def undeploy_machine(self, machine: Machine) -> None:
        """Undeploy a Kathara device.

        Args:
            machine (Kathara.model.Machine): A Kathara machine object.

        Returns:
            None

        Raises:
            LabNotFoundError: If the device specified is not associated to any network scenario.
        """
        if not machine.lab:
            raise LabNotFoundError(f"Device `{machine.name}` is not associated to a network scenario.")

        self.docker_machine.undeploy(machine.lab.hash, selected_machines={machine.name})
        self.docker_link.undeploy(machine.lab.hash, selected_links={x.link.name for x in machine.interfaces.values()})

    @privileged
    def undeploy_link(self, link: Link) -> None:
        """Undeploy a Kathara collision domain.

        Args:
            link (Kathara.model.Link): A Kathara collision domain object.

        Returns:
            None

        Raises:
            LabNotFoundError: If the collision domain is not associated to any network scenario.
        """
        if not link.lab:
            raise LabNotFoundError(f"Collision domain `{link.name}` is not associated to a network scenario.")

        self.docker_link.undeploy(link.lab.hash, selected_links={link.name})

    @privileged
    def undeploy_lab(self, lab_hash: Optional[str] = None, lab_name: Optional[str] = None, lab: Optional[Lab] = None,
                     selected_machines: Optional[Set[str]] = None,
                     excluded_machines: Optional[Set[str]] = None) -> None:
        """Undeploy a Kathara network scenario.

        Args:
            lab_hash (Optional[str]): The hash of the network scenario.
                Can be used as an alternative to lab_name and lab. If None, lab_name or lab should be set.
            lab_name (Optional[str]): The name of the network scenario.
                Can be used as an alternative to lab_hash and lab. If None, lab_hash or lab should be set.
            lab (Optional[Kathara.model.Lab]): The network scenario object.
                Can be used as an alternative to lab_hash and lab_name. If None, lab_hash or lab_name should be set.
            selected_machines (Optional[Set[str]]): If not None, undeploy only the specified devices.
            excluded_machines (Optional[Set[str]]): If not None, exclude devices from being undeployed.

        Returns:
            None

        Raises:
            InvocationError: If a running network scenario hash or name is not specified,
                or if both `selected_machines` and `excluded_machines` are specified.
        """
        check_required_single_not_none_var(lab_hash=lab_hash, lab_name=lab_name, lab=lab)
        if lab:
            lab_hash = lab.hash
        elif lab_name:
            lab_hash = utils.generate_urlsafe_hash(lab_name)

        if selected_machines and excluded_machines:
            raise InvocationError(f"You can either select or exclude devices.")

        self.docker_machine.undeploy(lab_hash, selected_machines=selected_machines, excluded_machines=excluded_machines)

        self.docker_link.undeploy(lab_hash)

    @privileged
    def wipe(self, all_users: bool = False) -> None:
        """Undeploy all the running network scenarios.

        If multiuser scenarios are active, undeploy only current user devices.

        Args:
            all_users (bool): If false, undeploy only the current user network scenarios. If true, undeploy the
                running network scenarios of all users.

        Returns:
            None
        """
        if Setting.get_instance().remote_url is not None and all_users:
            all_users = False
            logging.warning("Cannot wipe devices of other users with a remote Docker connection.")

        user_name = utils.get_current_user_name() if not all_users else None

        self.docker_machine.wipe(user=user_name)
        self.docker_link.wipe(user=user_name)

    @privileged
    def connect_tty(self, machine_name: str, lab_hash: Optional[str] = None, lab_name: Optional[str] = None,
                    lab: Optional[Lab] = None, shell: str = None, logs: bool = False,
                    wait: Union[bool, Tuple[int, float]] = True) -> None:
        """Connect to a device in a running network scenario, using the specified shell.

        Args:
            machine_name (str): The name of the device to connect.
            lab_hash (Optional[str]): The hash of the network scenario.
                Can be used as an alternative to lab_name and lab. If None, lab_name or lab should be set.
            lab_name (Optional[str]): The name of the network scenario.
                Can be used as an alternative to lab_hash and lab. If None, lab_hash or lab should be set.
            lab (Optional[Kathara.model.Lab]): The network scenario object.
                Can be used as an alternative to lab_hash and lab_name. If None, lab_hash or lab_name should be set.
            shell (str): The name of the shell to use for connecting.
            logs (bool): If True, print startup logs on stdout.
            wait (Union[bool, Tuple[int, float]]): If True, wait indefinitely until the end of the startup commands
                execution before connecting. If a tuple is provided, the first value indicates the number of retries
                before stopping waiting and the second value indicates the time interval to wait for each retry.
                Default is True.

        Returns:
            None

        Raises:
            InvocationError: If a running network scenario hash or name is not specified.
        """
        check_required_single_not_none_var(lab_hash=lab_hash, lab_name=lab_name, lab=lab)
        if lab:
            lab_hash = lab.hash
        elif lab_name:
            lab_hash = utils.generate_urlsafe_hash(lab_name)

        user_name = utils.get_current_user_name()

        self.docker_machine.connect(lab_hash=lab_hash,
                                    machine_name=machine_name,
                                    user=user_name,
                                    shell=shell,
                                    logs=logs,
                                    wait=wait
                                    )

    def connect_tty_obj(self, machine: Machine, shell: str = None, logs: bool = False,
                        wait: Union[bool, Tuple[int, float]] = True) -> None:
        """Connect to a device in a running network scenario, using the specified shell.

        Args:
            machine (Machine): The device to connect.
            shell (str): The name of the shell to use for connecting.
            logs (bool): If True, print startup logs on stdout.
            wait (Union[bool, Tuple[int, float]]): If True, wait indefinitely until the end of the startup commands
                execution before connecting. If a tuple is provided, the first value indicates the number of retries
                before stopping waiting and the second value indicates the time interval to wait for each retry.
                Default is True.

        Returns:
            None

        Raises:
            LabNotFoundError: If the specified device is not associated to any network scenario.
        """
        if not machine.lab:
            raise LabNotFoundError(f"Device `{machine.name}` is not associated to a network scenario.")

        self.connect_tty(machine.name, lab=machine.lab, shell=shell, logs=logs, wait=wait)

    @privileged
    def exec(self, machine_name: str, command: Union[List[str], str], lab_hash: Optional[str] = None,
             lab_name: Optional[str] = None, lab: Optional[Lab] = None, wait: Union[bool, Tuple[int, float]] = False,
             stream: bool = True) -> Union[DockerExecStream, Tuple[bytes, bytes, int]]:
        """Exec a command on a device in a running network scenario.

        Args:
            machine_name (str): The name of the device to connect.
            command (Union[List[str], str]): The command to exec on the device.
            lab_hash (Optional[str]): The hash of the network scenario.
                Can be used as an alternative to lab_name and lab. If None, lab_name or lab should be set.
            lab_name (Optional[str]): The name of the network scenario.
                Can be used as an alternative to lab_hash and lab. If None, lab_hash or lab should be set.
            lab (Optional[Kathara.model.Lab]): The network scenario object.
                Can be used as an alternative to lab_hash and lab_name. If None, lab_hash or lab_name should be set.
            wait (Union[bool, Tuple[int, float]]): If True, wait indefinitely until the end of the startup commands
                execution before executing the command. If a tuple is provided, the first value indicates the
                number of retries before stopping waiting and the second value indicates the time interval to wait
                for each retry. Default is False.
            stream (bool): If True, return a DockerExecStream object. If False,
                returns a tuple containing the complete stdout, the stderr, and the return code of the command.

        Returns:
            Union[DockerExecStream, Tuple[bytes, bytes, int]]: A DockerExecStream object or
            a tuple containing the stdout, the stderr and the return code of the command.

        Raises:
            InvocationError: If a running network scenario hash or name is not specified.
            MachineNotRunningError: If the specified device is not running.
            ValueError: If the wait values is neither a boolean nor a tuple, or an invalid tuple.
        """
        check_required_single_not_none_var(lab_hash=lab_hash, lab_name=lab_name, lab=lab)
        if lab:
            lab_hash = lab.hash
        elif lab_name:
            lab_hash = utils.generate_urlsafe_hash(lab_name)

        user_name = utils.get_current_user_name()
        return self.docker_machine.exec(
            lab_hash, machine_name, command, user=user_name, tty=False, wait=wait, stream=stream
        )

    def exec_obj(self, machine: Machine, command: Union[List[str], str], wait: Union[bool, Tuple[int, float]] = False,
                 stream: bool = True) -> Union[DockerExecStream, Tuple[bytes, bytes, int]]:
        """Exec a command on a device in a running network scenario.

        Args:
            machine (Machine): The device to connect.
            command (Union[List[str], str]): The command to exec on the device.
            wait (Union[bool, Tuple[int, float]]): If True, wait indefinitely until the end of the startup commands
                execution before executing the command. If a tuple is provided, the first value indicates the
                number of retries before stopping waiting and the second value indicates the time interval to wait
                for each retry. Default is False.
            stream (bool): If True, return a DockerExecStream object. If False,
                returns a tuple containing the complete stdout, the stderr, and the return code of the command.

        Returns:
            Union[DockerExecStream, Tuple[bytes, bytes, int]]: A DockerExecStream object or
            a tuple containing the stdout, the stderr and the return code of the command.

        Raises:
            LabNotFoundError: If the specified device is not associated to any network scenario.
            MachineNotRunningError: If the specified device is not running.
            MachineBinaryError: If the binary of the command is not found.
            ValueError: If the wait values is neither a boolean nor a tuple, or an invalid tuple.
        """
        if not machine.lab:
            raise LabNotFoundError(f"Device `{machine.name}` is not associated to a network scenario.")

        return self.exec(machine.name, command, lab=machine.lab, wait=wait, stream=stream)

    @privileged
    def copy_files(self, machine: Machine, guest_to_host: Dict[str, Union[str, io.IOBase]]) -> None:
        """Copy files on a running device in the specified paths.

        Args:
            machine (Kathara.model.Machine): A running device object. It must have the api_object field populated.
            guest_to_host (Dict[str, Union[str, io.IOBase]]): A dict containing the device path as key and a
                fileobj to copy in path as value or a path to a file.

        Returns:
            None
        """
        tar_data = pack_files_for_tar(guest_to_host)

        self.docker_machine.copy_files(machine.api_object,
                                       path="/",
                                       tar_data=tar_data
                                       )

    @privileged
    def retrieve_files(self, machine: Machine, src: str, dst: str) -> None:
        """Copy files from a running device path to the host.

        Args:
            machine (Kathara.model.Machine): A running device object. It must have the api_object field populated.
            src (str): The path of the file or folder to copy from the device.
            dst (str): The destination path on the host.

        Returns:
            None
        """
        self.docker_machine.retrieve_files(machine.api_object, src, dst)

    @privileged
    def get_machine_api_object(self, machine_name: str, lab_hash: Optional[str] = None, lab_name: Optional[str] = None,
                               lab: Optional[Lab] = None, all_users: bool = False) \
            -> docker.models.containers.Container:
        """Return the corresponding API object of a running device in a network scenario.

        Args:
            machine_name (str): The name of the device.
            lab_hash (Optional[str]): The hash of the network scenario.
                Can be used as an alternative to lab_name and lab. If None, lab_name or lab should be set.
            lab_name (Optional[str]): The name of the network scenario.
                Can be used as an alternative to lab_hash and lab. If None, lab_hash or lab should be set.
            lab (Optional[Kathara.model.Lab]): The network scenario object.
                Can be used as an alternative to lab_hash and lab_name. If None, lab_hash or lab_name should be set.
            all_users (bool): If True, return information about devices of all users.

        Returns:
            docker.models.containers.Container: Docker API object of devices.

        Raises:
            InvocationError: If a running network scenario hash or name is not specified.
            MachineNotFoundError: If the specified device is not found.
        """
        check_required_single_not_none_var(lab_hash=lab_hash, lab_name=lab_name, lab=lab)
        if lab:
            lab_hash = lab.hash
        elif lab_name:
            lab_hash = utils.generate_urlsafe_hash(lab_name)

        user_name = utils.get_current_user_name() if not all_users else None
        containers = self.docker_machine.get_machines_api_objects_by_filters(
            lab_hash=lab_hash, machine_name=machine_name, user=user_name
        )
        if containers:
            return containers.pop()

        raise MachineNotFoundError(f"Device `{machine_name}` not found.")

    @privileged
    def get_machines_api_objects(self, lab_hash: Optional[str] = None, lab_name: Optional[str] = None,
                                 lab: Optional[Lab] = None, all_users: bool = False) \
            -> List[docker.models.containers.Container]:
        """Return API objects of running devices.

        Args:
            lab_hash (Optional[str]): The hash of the network scenario.
                Can be used as an alternative to lab_name and lab.
            lab_name (Optional[str]): The name of the network scenario.
                Can be used as an alternative to lab_hash and lab.
            lab (Optional[Kathara.model.Lab]): The network scenario object.
                Can be used as an alternative to lab_hash and lab_name.
            all_users (bool): If True, return information about devices of all users.

        Returns:
            List[docker.models.containers.Container]: Docker API objects of devices.

        Raises:
            InvocationError: If a running network scenario hash or name is not specified.
        """
        check_single_not_none_var(lab_hash=lab_hash, lab_name=lab_name, lab=lab)
        if lab:
            lab_hash = lab.hash
        elif lab_name:
            lab_hash = utils.generate_urlsafe_hash(lab_name)

        user_name = utils.get_current_user_name() if not all_users else None
        return self.docker_machine.get_machines_api_objects_by_filters(lab_hash=lab_hash, user=user_name)

    @privileged
    def get_link_api_object(self, link_name: str, lab_hash: Optional[str] = None, lab_name: Optional[str] = None,
                            lab: Optional[Lab] = None, all_users: bool = False) -> docker.models.networks.Network:
        """Return the corresponding API object of a collision domain in a network scenario.

        Args:
            link_name (str): The name of the collision domain.
            lab_hash (Optional[str]): The hash of the network scenario.
                Can be used as an alternative to lab_name and lab. If None, lab_name or lab should be set.
            lab_name (Optional[str]): The name of the network scenario.
                Can be used as an alternative to lab_hash and lab. If None, lab_hash or lab should be set.
            lab (Optional[Kathara.model.Lab]): The network scenario object.
                Can be used as an alternative to lab_hash and lab_name. If None, lab_hash or lab_name should be set.
            all_users (bool): If True, return information about collision domains of all users.

        Returns:
            docker.models.networks.Network: Docker API object of the network.

        Raises:
            InvocationError: If a running network scenario hash or name is not specified.
            LinkNotFoundError: If the collision domain is not found.
        """
        check_required_single_not_none_var(lab_hash=lab_hash, lab_name=lab_name, lab=lab)
        if lab:
            lab_hash = lab.hash
        elif lab_name:
            lab_hash = utils.generate_urlsafe_hash(lab_name)

        user_name = utils.get_current_user_name() if not all_users else None
        networks = self.docker_link.get_links_api_objects_by_filters(
            lab_hash=lab_hash, link_name=link_name, user=user_name
        )
        if networks:
            return networks.pop()

        raise LinkNotFoundError(f"Collision Domain `{link_name}` not found.")

    @privileged
    def get_links_api_objects(self, lab_hash: Optional[str] = None, lab_name: Optional[str] = None,
                              lab: Optional[Lab] = None, all_users: bool = False) \
            -> List[docker.models.networks.Network]:
        """Return API objects of collision domains in a network scenario.

        Args:
            lab_hash (Optional[str]): The hash of the network scenario.
                Can be used as an alternative to lab_name and lab.
            lab_name (Optional[str]): The name of the network scenario.
                Can be used as an alternative to lab_hash and lab.
            lab (Optional[Kathara.model.Lab]): The network scenario object.
                Can be used as an alternative to lab_hash and lab_name.
            all_users (bool): If True, return information about collision domains of all users.

        Returns:
            List[docker.models.networks.Network]: Docker API objects of networks.

        Raises:
            InvocationError: If a running network scenario hash or name is not specified.
        """
        check_single_not_none_var(lab_hash=lab_hash, lab_name=lab_name, lab=lab)
        if lab:
            lab_hash = lab.hash
        elif lab_name:
            lab_hash = utils.generate_urlsafe_hash(lab_name)

        user_name = utils.get_current_user_name() if not all_users else None
        return self.docker_link.get_links_api_objects_by_filters(lab_hash=lab_hash, user=user_name)

    @privileged
    def get_lab_from_api(self, lab_hash: Optional[str] = None, lab_name: Optional[str] = None) -> Lab:
        """Return the network scenario (specified by the hash or name), building it from API objects.

        Args:
            lab_hash (Optional[str]): The hash of the network scenario.
                Can be used as an alternative to lab_name. If None, lab_name should be set.
            lab_name (Optional[str]): The name of the network scenario.
                Can be used as an alternative to lab_hash. If None, lab_hash should be set.

        Returns:
            Lab: The built network scenario.

        Raises:
            InvocationError: If a running network scenario hash or name is not specified.
        """
        if not lab_hash and not lab_name:
            raise InvocationError("You must specify a running network scenario hash or name.")

        if lab_name:
            reconstructed_lab = Lab(lab_name)
        else:
            reconstructed_lab = Lab("reconstructed_lab")
            reconstructed_lab.hash = lab_hash

        lab_containers = self.get_machines_api_objects(lab_hash=reconstructed_lab.hash)
        lab_networks = dict(
            map(lambda x: (x.name, x), self.get_links_api_objects(
                lab_hash=reconstructed_lab.hash \
                    if Setting.get_instance().shared_cds == SharedCollisionDomainsOption.NOT_SHARED else None,
                all_users=Setting.get_instance().shared_cds == SharedCollisionDomainsOption.USERS
            ))
        )

        for container in lab_containers:
            container.reload()
            device = reconstructed_lab.get_or_new_machine(container.labels["name"])
            device.api_object = container

            # Rebuild device metas
            # NOTE: We cannot rebuild "exec", "ipv6" and "num_terms" meta.
            device.add_meta("privileged", container.attrs["HostConfig"]["Privileged"])
            device.add_meta("image", container.attrs["Config"]["Image"])
            device.add_meta("shell", container.attrs["Config"]["Labels"]["shell"])

            # Memory is always returned in MBytes
            if container.attrs['HostConfig']['Memory'] > 0:
                device.add_meta("mem", f"{int(container.attrs['HostConfig']['Memory'] / (1024 ** 2))}M")

            # Reconvert nanocpus to a value passed by the user
            if container.attrs["HostConfig"]["NanoCpus"] > 0:
                device.add_meta("cpu", container.attrs["HostConfig"]["NanoCpus"] / 1000000000)

            for env in container.attrs["Config"]["Env"]:
                device.add_meta("env", env)

            # Reconvert ports to the device format
            if container.attrs['HostConfig']['PortBindings']:
                for port_info, port_data in container.attrs['HostConfig']['PortBindings'].items():
                    (guest_port, protocol) = port_info.split('/')
                    host_port = port_data[0]["HostPort"]
                    device.meta["ports"][(int(host_port), protocol)] = int(guest_port)

            # Reassign sysctls directly
            device.meta["sysctls"] = container.attrs["HostConfig"]["Sysctls"]

            if "none" not in container.attrs["NetworkSettings"]["Networks"]:
                if "bridge" in container.attrs["NetworkSettings"]["Networks"].keys():
                    device.add_meta("bridged", True)
                    device.add_meta("bridged_iface", int(container.labels['bridged_iface']))
                    container.attrs["NetworkSettings"]["Networks"].pop("bridge")

                networks = sorted(container.attrs["NetworkSettings"]["Networks"].items(),
                                  key=lambda x: x[1]["DriverOpts"]["kathara.iface"])

                for network_name, network_options in networks:
                    network = lab_networks[network_name]
                    link = reconstructed_lab.get_or_new_link(network.attrs["Labels"]["name"])
                    link.api_object = network
                    iface_number = int(network_options["DriverOpts"]["kathara.iface"])

                    iface_mac_addr = None
                    if network_options["DriverOpts"] is not None:
                        if "kathara.mac_addr" in network_options["DriverOpts"]:
                            iface_mac_addr = network_options["DriverOpts"]["kathara.mac_addr"]
                        if "com.docker.network.endpoint.sysctls" in network_options["DriverOpts"]:
                            for s in network_options["DriverOpts"]["com.docker.network.endpoint.sysctls"].split(","):
                                device.add_meta("sysctl", s.replace("IFNAME", f"eth{iface_number}"))

                    device.add_interface(link, mac_address=iface_mac_addr, number=iface_number)

        return reconstructed_lab

    @privileged
    def update_lab_from_api(self, lab: Lab) -> None:
        """Update the passed network scenario from API objects.

        Args:
            lab (Lab): The network scenario to update.
        """
        running_containers = self.get_machines_api_objects(lab_hash=lab.hash)

        deployed_networks = dict(
            map(lambda x: (x.name, x), self.get_links_api_objects(
                lab_hash=lab.hash \
                    if Setting.get_instance().shared_cds == SharedCollisionDomainsOption.NOT_SHARED else None,
                all_users=Setting.get_instance().shared_cds == SharedCollisionDomainsOption.USERS
            ))
        )
        for network in deployed_networks.values():
            network.reload()

        deployed_networks_by_link_name = dict(
            map(lambda x: (x.attrs["Labels"]["name"], x), deployed_networks.values())
        )

        for container in running_containers:
            container.reload()
            device = lab.get_or_new_machine(container.labels["name"])
            device.api_object = container

            # Collision domains declared in the network scenario
            static_links = set([x.link for x in device.interfaces.values()])
            # Interfaces currently attached to the device
            if "bridge" in container.attrs["NetworkSettings"]["Networks"].keys():
                container.attrs["NetworkSettings"]["Networks"].pop("bridge")

            if "none" in container.attrs["NetworkSettings"]["Networks"].keys():
                container.attrs["NetworkSettings"]["Networks"].pop("none")

            current_ifaces = [
                (lab.get_or_new_link(deployed_networks[name].attrs["Labels"]["name"]), options)
                for name, options in sorted(container.attrs["NetworkSettings"]["Networks"].items(),
                                            key=lambda x: x[1]["DriverOpts"]["kathara.iface"])
            ]

            # Collision domains currently attached to the device
            current_links = set(map(lambda x: x[0], current_ifaces))
            # Collision domains attached at runtime to the device
            dynamic_links = current_links - static_links
            # Static collision domains detached at runtime from the device
            deleted_links = static_links - current_links

            for link in static_links:
                if link.name in deployed_networks_by_link_name:
                    link.api_object = deployed_networks_by_link_name[link.name]

            current_ifaces = dict([(x[0].name, x[1]) for x in current_ifaces])
            for link in dynamic_links:
                link.api_object = deployed_networks_by_link_name[link.name]
                iface_options = current_ifaces[link.name]
                iface_mac_addr = None
                iface_number = int(iface_options["DriverOpts"]["kathara.iface"])

                if iface_options["DriverOpts"] is not None:
                    if "kathara.mac_addr" in iface_options["DriverOpts"]:
                        iface_mac_addr = iface_options["DriverOpts"]["kathara.mac_addr"]
                    if "com.docker.network.endpoint.sysctls" in iface_options["DriverOpts"]:
                        for s in iface_options["DriverOpts"]["com.docker.network.endpoint.sysctls"].split(","):
                            device.add_meta("sysctl", s.replace("IFNAME", f"eth{iface_number}"))

                device.add_interface(link, mac_address=iface_mac_addr, number=iface_number)

            for link in deleted_links:
                device.remove_interface(link)

    @privileged
    def get_machines_stats(self, lab_hash: Optional[str] = None, lab_name: Optional[str] = None,
                           lab: Optional[Lab] = None, machine_name: str = None, all_users: bool = False) \
            -> Generator[Dict[str, DockerMachineStats], None, None]:
        """Return information about the running devices.

        Args:
            lab_hash (Optional[str]): The hash of the network scenario.
                Can be used as an alternative to lab_name and lab.
            lab_name (Optional[str]): The name of the network scenario.
                Can be used as an alternative to lab_hash and lab.
            lab (Optional[Kathara.model.Lab]): The network scenario object.
                Can be used as an alternative to lab_hash and lab_name.
            machine_name (str): If specified return all the devices with machine_name.
            all_users (bool): If True, return information about the device of all users.

        Returns:
              Generator[Dict[str, DockerMachineStats], None, None]: A generator containing dicts that has API Object
              identifier as keys and DockerMachineStats objects as values.

        Raises:
            InvocationError: If more than one param among lab_hash, lab_name and lab is specified.
            PrivilegeError: If all_users is True and the user does not have root privileges.
        """
        check_single_not_none_var(lab_hash=lab_hash, lab_name=lab_name, lab=lab)

        if lab:
            lab_hash = lab.hash
        elif lab_name:
            lab_hash = utils.generate_urlsafe_hash(lab_name)

        user_name = utils.get_current_user_name() if not all_users else None
        return self.docker_machine.get_machines_stats(lab_hash=lab_hash, machine_name=machine_name,
                                                      user=user_name)

    @privileged
    def get_machine_stats(self, machine_name: str, lab_hash: Optional[str] = None, lab_name: Optional[str] = None,
                          lab: Optional[Lab] = None, all_users: bool = False) \
            -> Generator[Optional[DockerMachineStats], None, None]:
        """Return information of the specified device in a specified network scenario.

         Args:
            machine_name (str): The name of the device for which statistics are requested.
            lab_hash (Optional[str]): The hash of the network scenario.
                Can be used as an alternative to lab_name and lab. If None, lab_name or lab should be set.
            lab_name (Optional[str]): The name of the network scenario.
                Can be used as an alternative to lab_hash and lab. If None, lab_hash or lab should be set.
            lab (Optional[Kathara.model.Lab]): The network scenario object.
                Can be used as an alternative to lab_hash and lab_name. If None, lab_hash or lab_name should be set.
            all_users (bool): If True, search the device among all the users devices.

        Returns:
            Generator[Optional[DockerMachineStats], None, None]: A generator containing the DockerMachineStats object
            with the device info. Returns None if the device is not found.

        Raises:
            InvocationError: If a running network scenario hash, name or object is not specified.
            PrivilegeError: If all_users is True and the user does not have root privileges.
        """
        check_required_single_not_none_var(lab_hash=lab_hash, lab_name=lab_name, lab=lab)

        machines_stats = self.get_machines_stats(lab_hash=lab_hash, lab_name=lab_name, lab=lab,
                                                 machine_name=machine_name, all_users=all_users)
        machines_stats_next = next(machines_stats)
        if machines_stats_next:
            (_, machine_stats) = machines_stats_next.popitem()
            yield machine_stats
        else:
            yield None

    def get_machine_stats_obj(self, machine: Machine, all_users: bool = False) \
            -> Generator[Optional[DockerMachineStats], None, None]:
        """Return information of the specified device in a specified network scenario.

        Args:
            machine (Machine): The device for which statistics are requested.
            all_users (bool): If True, search the device among all the users devices.

        Returns:
            Generator[Optional[IMachineStats], None, None]: A generator containing the IMachineStats object
            with the device info. Returns None if the device is not found.

        Raises:
            LabNotFoundError: If the specified device is not associated to any network scenario.
            MachineNotRunningError: If the specified device is not running.
            PrivilegeError: If all_users is True and the user does not have root privileges.
        """
        if not machine.lab:
            raise LabNotFoundError("Device `%s` is not associated to a network scenario." % machine.name)

        return self.get_machine_stats(machine.name, lab=machine.lab, all_users=all_users)

    @privileged
    def get_links_stats(self, lab_hash: Optional[str] = None, lab_name: Optional[str] = None, lab: Optional[Lab] = None,
                        link_name: str = None, all_users: bool = False) \
            -> Generator[Dict[str, DockerLinkStats], None, None]:
        """Return information about deployed networks.

        Args:
            lab_hash (Optional[str]): The hash of the network scenario.
                Can be used as an alternative to lab_name and lab.
            lab_name (Optional[str]): The name of the network scenario.
                Can be used as an alternative to lab_hash and lab.
            lab (Optional[Kathara.model.Lab]): The network scenario object.
                Can be used as an alternative to lab_hash and lab_name.
           link_name (str): If specified return all the networks with link_name.
           all_users (bool): If True, return information about the networks of all users.

        Returns:
             Generator[Dict[str, DockerLinkStats], None, None]: A generator containing dicts that has API Object
                identifier as keys and DockerLinksStats objects as values.

        Raises:
            InvocationError: If a running network scenario hash, name or object is not specified.
            PrivilegeError: If all_users is True and the user does not have root privileges.
        """
        check_single_not_none_var(lab_hash=lab_hash, lab_name=lab_name, lab=lab)
        if lab:
            lab_hash = lab.hash
        elif lab_name:
            lab_hash = utils.generate_urlsafe_hash(lab_name)

        user_name = utils.get_current_user_name() if not all_users else None
        return self.docker_link.get_links_stats(lab_hash=lab_hash, link_name=link_name, user=user_name)

    @privileged
    def get_link_stats(self, link_name: str, lab_hash: Optional[str] = None, lab_name: Optional[str] = None,
                       lab: Optional[Lab] = None, all_users: bool = False) \
            -> Generator[Optional[DockerLinkStats], None, None]:
        """Return information of the specified deployed network in a specified network scenario.

        Args:
            link_name (str): The name of the collision domain for which statistics are requested.
            lab_hash (Optional[str]): The hash of the network scenario.
                Can be used as an alternative to lab_name and lab. If None, lab_name or lab should be set.
            lab_name (Optional[str]): The name of the network scenario.
                Can be used as an alternative to lab_hash and lab. If None, lab_hash or lab should be set.
            lab (Optional[Kathara.model.Lab]): The network scenario object.
                Can be used as an alternative to lab_hash and lab_name. If None, lab_hash or lab_name should be set.
            all_users (bool): If True, return information about the networks of all users.

        Returns:
            Generator[Optional[DockerLinkStats], None, None]: A generator containing the DockerLinkStats object
            with the network info. Returns None if the network is not found.

        Raises:
            InvocationError: If a running network scenario hash or name is not specified.
            PrivilegeError: If all_users is True and the user does not have root privileges.
        """
        check_required_single_not_none_var(lab_hash=lab_hash, lab_name=lab_name, lab=lab)
        if lab:
            lab_hash = lab.hash
        elif lab_name:
            lab_hash = utils.generate_urlsafe_hash(lab_name)

        links_stats = self.get_links_stats(lab_hash=lab_hash, link_name=link_name, all_users=all_users)
        links_stats_next = next(links_stats)
        if links_stats_next:
            (_, link_stats) = links_stats_next.popitem()
            yield link_stats
        else:
            yield None

    def get_link_stats_obj(self, link: Link, all_users: bool = False) \
            -> Generator[Optional[DockerLinkStats], None, None]:
        """Return information of the specified deployed network in a specified network scenario.

        Args:
            link (Link): The collision domain for which statistics are requested.
            all_users (bool): If True, return information about the networks of all users.

        Returns:
            Generator[Optional[ILinkStats], None, None]: A generator containing the ILinkStats object
            with the network info. Returns None if the network is not found.

        Raises:
            LabNotFoundError: If the specified device is not associated to any network scenario.
            PrivilegeError: If all_users is True and the user does not have root privileges.
        """
        if not link.lab:
            raise LabNotFoundError(f"Link `{link.name}` is not associated to a network scenario.")

        return self.get_link_stats(link.name, lab=link.lab, all_users=all_users)

    @privileged
    def check_image(self, image_name: str) -> None:
        """Check if the specified image is valid.

        Args:
            image_name (str): The name of the image

        Returns:
            None

        Raises:
            ConnectionError: If the image is not locally available and there is no connection to a remote image repository.
            DockerImageNotFoundError: If the image is not found.
        """
        self.docker_image.check(image_name)

    @privileged
    def get_release_version(self) -> str:
        """Return the current manager version.

        Returns:
            str: The current manager version.
        """
        return self.client.version()["Version"]

    @staticmethod
    def get_formatted_manager_name() -> str:
        """Return a formatted string containing the current manager name.

        Returns:
            str: A formatted string containing the current manager name.
        """
        return "Docker (Kathara)"
